import time
from concurrent.futures import ThreadPoolExecutor, wait

from palimpzest.constants import PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS, OptimizationStrategy
from palimpzest.core.data.dataclasses import ExecutionStats, OperatorStats, PlanStats
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.lib.schemas import SourceRecord
from palimpzest.policy import Policy
from palimpzest.query.execution.execution_engine import ExecutionEngine
from palimpzest.query.execution.plan_executors.parallel_plan_execution import (
    PipelinedParallelPlanExecutor,
)
from palimpzest.query.execution.plan_executors.single_threaded_plan_execution import (
    PipelinedSingleThreadPlanExecutor,
    SequentialSingleThreadPlanExecutor,
)
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.datasource import DataSourcePhysicalOp, MarshalAndScanDataOp
from palimpzest.query.operators.filter import FilterOp
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.optimizer.cost_model import CostModel
from palimpzest.query.optimizer.optimizer import Optimizer
from palimpzest.query.optimizer.plan import PhysicalPlan
from palimpzest.sets import Set
from palimpzest.utils.progress import create_progress_manager


class NoSentinelExecutionEngine(ExecutionEngine):
    """
    This class implements the abstract execute() method from the ExecutionEngine.
    This class still needs to be sub-classed by another Execution class which implements
    the execute_plan() method.
    """

    def execute(self, dataset: Set, policy: Policy):
        execution_start_time = time.time()

        # if nocache is True, make sure we do not re-use codegen examples
        if self.nocache:
            self.clear_cached_examples()

        # construct the CostModel
        cost_model = CostModel()

        # initialize the optimizer
        optimizer = Optimizer(
            policy=policy,
            cost_model=cost_model,
            no_cache=self.nocache,
            verbose=self.verbose,
            available_models=self.available_models,
            allow_bonded_query=self.allow_bonded_query,
            allow_conventional_query=self.allow_conventional_query,
            allow_code_synth=self.allow_code_synth,
            allow_token_reduction=self.allow_token_reduction,
            allow_rag_reduction=self.allow_rag_reduction,
            allow_mixtures=self.allow_mixtures,
            optimization_strategy=self.optimization_strategy,
        )

        # execute plan(s) according to the optimization strategy
        records, plan_stats = [], []
        if self.optimization_strategy == OptimizationStrategy.CONFIDENCE_INTERVAL:
            records, plan_stats = self.execute_confidence_interval_strategy(dataset, policy, optimizer)
        
        else:
            records, plan_stats = self.execute_strategy(dataset, policy, optimizer)

        # aggregate plan stats
        aggregate_plan_stats = self.aggregate_plan_stats(plan_stats)

        # add sentinel records and plan stats (if captured) to plan execution data
        execution_stats = ExecutionStats(
            execution_id=self.execution_id(),
            plan_stats=aggregate_plan_stats,
            total_execution_time=time.time() - execution_start_time,
            total_execution_cost=sum(
                list(map(lambda plan_stats: plan_stats.total_plan_cost, aggregate_plan_stats.values()))
            ),
            plan_strs={plan_id: plan_stats.plan_str for plan_id, plan_stats in aggregate_plan_stats.items()},
        )

        return records, execution_stats


class NoSentinelSequentialSingleThreadExecution(NoSentinelExecutionEngine, SequentialSingleThreadPlanExecutor):
    """
    This class performs non-sample based execution while executing plans in a sequential, single-threaded fashion.
    """
    def __init__(self, *args, **kwargs):
        NoSentinelExecutionEngine.__init__(self, *args, **kwargs)
        SequentialSingleThreadPlanExecutor.__init__(self, *args, **kwargs)
        self.progress_manager = None

    def execute_plan(self, plan: PhysicalPlan, num_samples: int | float = float("inf"), plan_workers: int = 1):
        """Initialize the stats and execute the plan with progress reporting."""
        if self.verbose:
            print("----------------------")
            print(f"PLAN[{plan.plan_id}] (n={num_samples}):")
            print(plan)
            print("---")

        plan_start_time = time.time()

        # Initialize progress manager
        self.progress_manager = create_progress_manager()

        # initialize plan stats and operator stats
        plan_stats = PlanStats(plan_id=plan.plan_id, plan_str=str(plan))
        for op in plan.operators:
            op_id = op.get_op_id()
            op_name = op.op_name()
            op_details = {k: str(v) for k, v in op.get_id_params().items()}
            plan_stats.operator_stats[op_id] = OperatorStats(op_id=op_id, op_name=op_name, op_details=op_details)

        # initialize list of output records and intermediate variables
        output_records = []
        current_scan_idx = self.scan_start_idx

        # get handle to DataSource and pre-compute its size
        source_operator = plan.operators[0]
        datasource = (
            self.datadir.get_registered_dataset(source_operator.dataset_id)
            if isinstance(source_operator, MarshalAndScanDataOp)
            else self.datadir.get_cached_result(source_operator.dataset_id)
        )
        datasource_len = len(datasource)

        # Calculate total work units - each record needs to go through each operator
        total_ops = len(plan.operators)
        total_items = min(num_samples, datasource_len) if num_samples != float("inf") else datasource_len
        total_work_units = total_items * total_ops
        self.progress_manager.start(total_work_units)
        work_units_completed = 0

        # initialize processing queues for each operation
        processing_queues = {op.get_op_id(): [] for op in plan.operators if not isinstance(op, DataSourcePhysicalOp)}

        try:
            # execute the plan one operator at a time
            for op_idx, operator in enumerate(plan.operators):
                op_id = operator.get_op_id()
                prev_op_id = plan.operators[op_idx - 1].get_op_id() if op_idx > 1 else None
                next_op_id = plan.operators[op_idx + 1].get_op_id() if op_idx + 1 < len(plan.operators) else None

                # Update progress to show which operator is currently running
                op_name = operator.__class__.__name__
                self.progress_manager.update(work_units_completed, f"Running {op_name} ({op_idx + 1}/{total_ops})")

                # initialize output records and record_op_stats for this operator
                records, record_op_stats = [], []

                # invoke datasource operator(s) until we run out of source records or hit the num_samples limit
                if isinstance(operator, DataSourcePhysicalOp):
                    keep_scanning_source_records = True
                    while keep_scanning_source_records:
                        # construct input DataRecord for DataSourcePhysicalOp
                        # NOTE: this DataRecord will be discarded and replaced by the scan_operator;
                        #       it is simply a vessel to inform the scan_operator which record to fetch
                        candidate = DataRecord(schema=SourceRecord, source_id=current_scan_idx)
                        candidate.idx = current_scan_idx
                        candidate.get_item_fn = datasource.get_item

                        # run DataSourcePhysicalOp on record
                        record_set = operator(candidate)
                        records.extend(record_set.data_records)
                        record_op_stats.extend(record_set.record_op_stats)

                        # Update progress for each processed record in data source
                        work_units_completed += 1
                        self.progress_manager.update(
                            work_units_completed, 
                            f"Scanning data source: {current_scan_idx + 1}/{total_items}"
                        )

                        # update the current scan index
                        current_scan_idx += 1

                        # update whether to keep scanning source records
                        keep_scanning_source_records = current_scan_idx < datasource_len and len(records) < num_samples

                # aggregate operators accept all input records at once
                elif isinstance(operator, AggregateOp):
                    record_set = operator(candidates=processing_queues[op_id])
                    records = record_set.data_records
                    record_op_stats = record_set.record_op_stats
                    
                    # Update progress for aggregate operation - count all records being aggregated
                    work_units_completed += len(processing_queues[op_id])
                    self.progress_manager.update(
                        work_units_completed,
                        f"Aggregating {len(processing_queues[op_id])} records"
                    )

                # otherwise, process the records in the processing queue for this operator one at a time
                elif len(processing_queues[op_id]) > 0:
                    queue_size = len(processing_queues[op_id])
                    for idx, input_record in enumerate(processing_queues[op_id]):
                        record_set = operator(input_record)
                        records.extend(record_set.data_records)
                        record_op_stats.extend(record_set.record_op_stats)

                        # Update progress for each processed record in the queue
                        work_units_completed += 1
                        self.progress_manager.update(
                            work_units_completed,
                            f"Processing records: {idx + 1}/{queue_size}"
                        )

                        if isinstance(operator, LimitScanOp) and len(records) == operator.limit:
                            break

                # update plan stats
                plan_stats.operator_stats[op_id].add_record_op_stats(
                    record_op_stats,
                    source_op_id=prev_op_id,
                    plan_id=plan.plan_id,
                )

                # add records (which are not filtered) to the cache, if allowed
                if not self.nocache:
                    for record in records:
                        if getattr(record, "passed_operator", True):
                            if operator.target_cache_id is None:
                                print("No cache ID for operator", operator)
                                breakpoint()
                            self.datadir.append_cache(operator.target_cache_id, record)

                # update processing_queues or output_records
                for record in records:
                    if isinstance(operator, FilterOp) and not record.passed_operator:
                        continue
                    if next_op_id is not None:
                        processing_queues[next_op_id].append(record)
                    else:
                        output_records.append(record)

                # if we've filtered out all records, terminate early
                if next_op_id is not None and processing_queues[next_op_id] == []:
                    break

            # if caching was allowed, close the cache
            if not self.nocache:
                for operator in plan.operators:
                    self.datadir.close_cache(operator.target_cache_id)

            # finalize plan stats
            total_plan_time = time.time() - plan_start_time
            plan_stats.finalize(total_plan_time)

        finally:
            # Always finish progress tracking
            if self.progress_manager:
                self.progress_manager.finish()

        return output_records, plan_stats


class NoSentinelPipelinedSingleThreadExecution(NoSentinelExecutionEngine, PipelinedSingleThreadPlanExecutor):
    """
    This class performs non-sample based execution while executing plans in a pipelined, single-threaded fashion.
    """
    def __init__(self, *args, **kwargs):
        NoSentinelExecutionEngine.__init__(self, *args, **kwargs)
        PipelinedSingleThreadPlanExecutor.__init__(self, *args, **kwargs)
        self.progress_manager = None

    def execute_plan(self, plan: PhysicalPlan, num_samples: int | float = float("inf"), plan_workers: int = 1):
        """Initialize the stats and execute the plan with progress reporting."""
        if self.verbose:
            print("----------------------")
            print(f"PLAN[{plan.plan_id}] (n={num_samples}):")
            print(plan)
            print("---")

        plan_start_time = time.time()

        # Initialize progress manager
        self.progress_manager = create_progress_manager()

        # initialize plan stats and operator stats
        plan_stats = PlanStats(plan_id=plan.plan_id, plan_str=str(plan))
        for op in plan.operators:
            op_id = op.get_op_id()
            op_name = op.op_name()
            op_details = {k: str(v) for k, v in op.get_id_params().items()}
            plan_stats.operator_stats[op_id] = OperatorStats(op_id=op_id, op_name=op_name, op_details=op_details)

        # initialize list of output records and intermediate variables
        output_records = []
        source_records_scanned = 0
        current_scan_idx = self.scan_start_idx

        # get handle to DataSource and pre-compute its size
        source_operator = plan.operators[0]
        datasource = (
            self.datadir.get_registered_dataset(source_operator.dataset_id)
            if isinstance(source_operator, MarshalAndScanDataOp)
            else self.datadir.get_cached_result(source_operator.dataset_id)
        )
        datasource_len = len(datasource)

        # Calculate total work units - each record needs to go through each operator
        total_ops = len(plan.operators)
        total_items = min(num_samples, datasource_len) if num_samples != float("inf") else datasource_len
        total_work_units = total_items * total_ops
        self.progress_manager.start(total_work_units)
        work_units_completed = 0

        try:
            # initialize processing queues for each operation
            processing_queues = {op.get_op_id(): [] for op in plan.operators if not isinstance(op, DataSourcePhysicalOp)}

            # execute the plan until either:
            # 1. all records have been processed, or
            # 2. the final limit operation has completed
            finished_executing, keep_scanning_source_records = False, True
            while not finished_executing:
                for op_idx, operator in enumerate(plan.operators):
                    op_id = operator.get_op_id()
                    prev_op_id = plan.operators[op_idx - 1].get_op_id() if op_idx > 1 else None
                    next_op_id = plan.operators[op_idx + 1].get_op_id() if op_idx + 1 < len(plan.operators) else None

                    # Update progress with current operator info
                    op_name = operator.__class__.__name__
                    self.progress_manager.update(work_units_completed, f"Running {op_name} ({op_idx + 1}/{total_ops})")

                    # create empty lists for records and execution stats generated by executing this operator on its next input(s)
                    records, record_op_stats = [], []

                    # invoke datasource operator(s) until we run out of source records or hit the num_samples limit
                    if isinstance(operator, DataSourcePhysicalOp):
                        if keep_scanning_source_records:
                            # construct input DataRecord for DataSourcePhysicalOp
                            candidate = DataRecord(schema=SourceRecord, source_id=current_scan_idx)
                            candidate.idx = current_scan_idx
                            candidate.get_item_fn = datasource.get_item

                            # run DataSourcePhysicalOp on record
                            record_set = operator(candidate)
                            records = record_set.data_records
                            record_op_stats = record_set.record_op_stats

                            # Update progress for each processed record
                            work_units_completed += 1
                            self.progress_manager.update(
                                work_units_completed,
                                f"Scanning data source: {current_scan_idx + 1}/{total_items}"
                            )

                            # update number of source records scanned and the current index
                            source_records_scanned += len(records)
                            current_scan_idx += 1

                            # update whether to keep scanning source records
                            keep_scanning_source_records = current_scan_idx < datasource_len and source_records_scanned < num_samples

                    # only invoke aggregate operator(s) once there are no more source records and all
                    # upstream operators' processing queues are empty
                    elif isinstance(operator, AggregateOp):
                        upstream_ops_are_finished = True
                        for upstream_op_idx in range(op_idx):
                            # datasources do not have processing queues
                            if isinstance(plan.operators[upstream_op_idx], DataSourcePhysicalOp):
                                continue

                            # check upstream ops which do have a processing queue
                            upstream_op_id = plan.operators[upstream_op_idx].get_op_id()
                            upstream_ops_are_finished = (
                                upstream_ops_are_finished and len(processing_queues[upstream_op_id]) == 0
                            )

                        if not keep_scanning_source_records and upstream_ops_are_finished:
                            record_set = operator(candidates=processing_queues[op_id])
                            records = record_set.data_records
                            record_op_stats = record_set.record_op_stats
                            processing_queues[op_id] = []

                            # Update progress for aggregate operation
                            work_units_completed += len(processing_queues[op_id])
                            self.progress_manager.update(
                                work_units_completed,
                                f"Aggregating {len(processing_queues[op_id])} records"
                            )

                    # otherwise, process the next record in the processing queue for this operator
                    elif len(processing_queues[op_id]) > 0:
                        input_record = processing_queues[op_id].pop(0)
                        record_set = operator(input_record)
                        records = record_set.data_records
                        record_op_stats = record_set.record_op_stats

                        # Update progress for processed record
                        work_units_completed += 1
                        self.progress_manager.update(
                            work_units_completed,
                            f"Processing record through {op_name}"
                        )

                    # if records were generated by this operator, process them
                    if len(records) > 0:
                        # update plan stats
                        plan_stats.operator_stats[op_id].add_record_op_stats(
                            record_op_stats,
                            source_op_id=prev_op_id,
                            plan_id=plan.plan_id,
                        )

                        # add records (which are not filtered) to the cache, if allowed
                        if not self.nocache:
                            for record in records:
                                if getattr(record, "passed_operator", True):
                                    self.datadir.append_cache(operator.target_cache_id, record)

                        # update processing_queues or output_records
                        for record in records:
                            if isinstance(operator, FilterOp) and not record.passed_operator:
                                continue
                            if next_op_id is not None:
                                processing_queues[next_op_id].append(record)
                            else:
                                output_records.append(record)

                # update finished_executing based on whether all records have been processed
                still_processing = any([len(queue) > 0 for queue in processing_queues.values()])
                finished_executing = not keep_scanning_source_records and not still_processing

                # update finished_executing based on limit
                if isinstance(operator, LimitScanOp):
                    finished_executing = len(output_records) == operator.limit

            # if caching was allowed, close the cache
            if not self.nocache:
                for operator in plan.operators:
                    self.datadir.close_cache(operator.target_cache_id)

            # finalize plan stats
            total_plan_time = time.time() - plan_start_time
            plan_stats.finalize(total_plan_time)

        finally:
            # Always finish progress tracking
            if self.progress_manager:
                self.progress_manager.finish()

        return output_records, plan_stats


class NoSentinelPipelinedParallelExecution(NoSentinelExecutionEngine, PipelinedParallelPlanExecutor):
    """
    This class performs non-sample based execution while executing plans in a pipelined, parallel fashion.
    """
    def __init__(self, *args, **kwargs):
        NoSentinelExecutionEngine.__init__(self, *args, **kwargs)
        PipelinedParallelPlanExecutor.__init__(self, *args, **kwargs)
        self.progress_manager = None

    # def execute_plan(self, plan: PhysicalPlan, num_samples: int | float = float("inf"), plan_workers: int = 1):
    #     """Initialize the stats and execute the plan with progress reporting."""
    #     if self.verbose:
    #         print("----------------------")
    #         print(f"PLAN[{plan.plan_id}] (n={num_samples}):")
    #         print(plan)
    #         print("---")

    #     plan_start_time = time.time()

    #     # Initialize progress manager
    #     self.progress_manager = create_progress_manager()

    #     # initialize plan stats and operator stats
    #     plan_stats = PlanStats(plan_id=plan.plan_id, plan_str=str(plan))
    #     for op in plan.operators:
    #         op_id = op.get_op_id()
    #         op_name = op.op_name()
    #         op_details = {k: str(v) for k, v in op.get_id_params().items()}
    #         plan_stats.operator_stats[op_id] = OperatorStats(op_id=op_id, op_name=op_name, op_details=op_details)

    #     # initialize list of output records and intermediate variables
    #     output_records = []
    #     source_records_scanned = 0
    #     current_scan_idx = self.scan_start_idx

    #     # get handle to DataSource and pre-compute its size
    #     source_operator = plan.operators[0]
    #     datasource = (
    #         self.datadir.get_registered_dataset(source_operator.dataset_id)
    #         if isinstance(source_operator, MarshalAndScanDataOp)
    #         else self.datadir.get_cached_result(source_operator.dataset_id)
    #     )
    #     datasource_len = len(datasource)

    #     # Calculate total work units - each record needs to go through each operator
    #     total_ops = len(plan.operators)
    #     total_items = min(num_samples, datasource_len) if num_samples != float("inf") else datasource_len
    #     total_work_units = total_items * total_ops
    #     self.progress_manager.start(total_work_units)
    #     work_units_completed = 0

    #     try:
    #         with ThreadPoolExecutor(max_workers=plan_workers) as executor:
    #             # initialize processing queues and futures for each operation
    #             processing_queues = {op.get_op_id(): [] for op in plan.operators}
    #             futures = []

    #             # execute the plan until either:
    #             # 1. all records have been processed, or
    #             # 2. the final limit operation has completed
    #             finished_executing, keep_scanning_source_records = False, True
    #             last_work_units_completed = 0
    #             while not finished_executing:
    #                 # Process completed futures
    #                 done_futures, not_done_futures = wait(futures, timeout=PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS)
    #                 futures = list(not_done_futures)

    #                 for future in done_futures:
    #                     record_set, operator = future.result()
    #                     op_id = operator.get_op_id()
    #                     op_idx = next(i for i, op in enumerate(plan.operators) if op.get_op_id() == op_id)
    #                     next_op_id = plan.operators[op_idx + 1].get_op_id() if op_idx + 1 < len(plan.operators) else None

    #                     # Update progress for completed operation
    #                     work_units_completed += len(record_set.data_records)
    #                     if work_units_completed > last_work_units_completed:
    #                         self.progress_manager.update(
    #                             work_units_completed,
    #                             f"Completed {operator.__class__.__name__} on {len(record_set.data_records)} records"
    #                         )
    #                         last_work_units_completed = work_units_completed

    #                     # Process records
    #                     for record in record_set:
    #                         if isinstance(operator, FilterOp) and not record.passed_operator:
    #                             continue
    #                         if next_op_id is not None:
    #                             processing_queues[next_op_id].append(record)
    #                         else:
    #                             output_records.append(record)

    #                 # Submit new tasks
    #                 for _, operator in enumerate(plan.operators):
    #                     op_id = operator.get_op_id()
                        
    #                     if isinstance(operator, DataSourcePhysicalOp) and keep_scanning_source_records:
    #                         # Submit source operator task
    #                         candidate = DataRecord(schema=SourceRecord, source_id=current_scan_idx)
    #                         candidate.idx = current_scan_idx
    #                         candidate.get_item_fn = datasource.get_item
    #                         futures.append(executor.submit(self.execute_op_wrapper, operator, candidate))
    #                         current_scan_idx += 1
    #                         keep_scanning_source_records = current_scan_idx < datasource_len and source_records_scanned < num_samples
                        
    #                     elif len(processing_queues[op_id]) > 0:
    #                         # Submit task for next record in queue
    #                         input_record = processing_queues[op_id].pop(0)
    #                         futures.append(executor.submit(self.execute_op_wrapper, operator, input_record))

    #                 # Check if we're done
    #                 still_processing = any([len(queue) > 0 for queue in processing_queues.values()])
    #                 finished_executing = not keep_scanning_source_records and not still_processing and len(futures) == 0

    #         # if caching was allowed, close the cache
    #         if not self.nocache:
    #             for operator in plan.operators:
    #                 self.datadir.close_cache(operator.target_cache_id)

    #         # finalize plan stats
    #         total_plan_time = time.time() - plan_start_time
    #         plan_stats.finalize(total_plan_time)

    #     finally:
    #         # Always finish progress tracking
    #         if self.progress_manager:
    #             self.progress_manager.finish()

    #     return output_records, plan_stats

    



from palimpzest.query.execution.execute import Execute
from palimpzest.query.execution.mab_sentinel_execution import (
    MABSequentialParallelSentinelExecution,
    MABSequentialSingleThreadSentinelExecution,
)
from palimpzest.query.execution.nosentinel_execution import (
    NoSentinelPipelinedParallelExecution,
    NoSentinelPipelinedSingleThreadExecution,
    NoSentinelSequentialSingleThreadExecution,
)
from palimpzest.query.execution.random_sampling_sentinel_execution import (
    RandomSamplingSequentialParallelSentinelExecution,
    RandomSamplingSequentialSingleThreadSentinelExecution,
)
from palimpzest.query.execution.streaming_execution import StreamingSequentialExecution
from palimpzest.query.operators.aggregate import AggregateOp, ApplyGroupByOp, AverageAggregateOp, CountAggregateOp
from palimpzest.query.operators.convert import (
    ConvertOp,
    LLMConvert,
    LLMConvertBonded,
    LLMConvertConventional,
    NonLLMConvert,
)
from palimpzest.query.operators.datasource import CacheScanDataOp, DataSourcePhysicalOp, MarshalAndScanDataOp
from palimpzest.query.operators.filter import FilterOp, LLMFilter, NonLLMFilter
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.logical import (
    Aggregate,
    BaseScan,
    CacheScan,
    ConvertScan,
    FilteredScan,
    GroupByAggregate,
    LimitScan,
    LogicalOperator,
    Project,
    RetrieveScan,
)
from palimpzest.query.operators.physical import PhysicalOperator

__all__ = [
    # execute
    "Execute",
    # mab_sentinel_execution
    "MABSequentialParallelSentinelExecution",
    "MABSequentialSingleThreadSentinelExecution",
    # nosentinel_execution
    "NoSentinelPipelinedParallelExecution",
    "NoSentinelPipelinedSingleThreadExecution",
    "NoSentinelSequentialSingleThreadExecution",
    # random_sampling_sentinel_execution
    "RandomSamplingSequentialParallelSentinelExecution",
    "RandomSamplingSequentialSingleThreadSentinelExecution",
    # streaming_execution
    "StreamingSequentialExecution",
    # aggregate
    "AggregateOp",
    "ApplyGroupByOp",
    "AverageAggregateOp",
    "CountAggregateOp",
    # convert
    "ConvertOp",
    "LLMConvert",
    "LLMConvertBonded",
    "LLMConvertConventional",
    "NonLLMConvert",
    # datasource
    "CacheScanDataOp",
    "DataSourcePhysicalOp",
    "MarshalAndScanDataOp",
    # filter
    "FilterOp",
    "LLMFilter",
    "NonLLMFilter",
    # limit
    "LimitScanOp",
    # logical
    "Aggregate",
    "BaseScan",
    "CacheScan",
    "ConvertScan",
    "FilteredScan",
    "GroupByAggregate",
    "LimitScan",
    "LogicalOperator",
    "Project",
    "RetrieveScan",
    # physical
    "PhysicalOperator",
]


import time
from concurrent.futures import ThreadPoolExecutor, wait
from functools import partial
from typing import Callable

import numpy as np

from palimpzest.constants import PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS, OptimizationStrategy
from palimpzest.core.data.dataclasses import ExecutionStats, OperatorStats, PlanStats, RecordOpStats
from palimpzest.core.data.datasources import ValidationDataSource
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.schemas import SourceRecord
from palimpzest.policy import Policy
from palimpzest.query.execution.execution_engine import ExecutionEngine
from palimpzest.query.execution.plan_executors.single_threaded_plan_execution import SequentialSingleThreadPlanExecutor
from palimpzest.query.operators.convert import ConvertOp, LLMConvert
from palimpzest.query.operators.datasource import CacheScanDataOp, MarshalAndScanDataOp
from palimpzest.query.operators.filter import FilterOp, LLMFilter
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.retrieve import RetrieveOp
from palimpzest.query.optimizer.cost_model import CostModel, SampleBasedCostModel
from palimpzest.query.optimizer.optimizer import Optimizer
from palimpzest.query.optimizer.plan import SentinelPlan
from palimpzest.sets import Set


class RandomSamplingSentinelExecutionEngine(ExecutionEngine):
    """
    This class implements the abstract execute() method from the ExecutionEngine.
    This class still needs to be sub-classed by another Execution class which implements
    the higher-level execute_plan() method.
    """
    def __init__(
            self,
            k: int,
            sample_budget: int,
            sample_all_ops: bool = False,
            sample_all_records: bool = False,
            sample_start_idx: int | None = None,
            sample_end_idx: int | None = None,
            use_final_op_quality: bool = False,
            seed: int = 42,
            exp_name: str | None = None,
            *args,
            **kwargs,
        ):
        super().__init__(*args, **kwargs)
        # self.max_workers = self.get_parallel_max_workers()
        # TODO: undo
        # self.max_workers = 1
        assert isinstance(self.datasource, ValidationDataSource), "DataSource must be ValidationDataSource for sentinel execution"

        self.k = k
        self.sample_budget = sample_budget
        self.j = int(sample_budget / k)
        self.sample_all_ops = sample_all_ops
        self.sample_all_records = sample_all_records
        self.sample_start_idx = sample_start_idx
        self.sample_end_idx = sample_end_idx
        self.use_final_op_quality = use_final_op_quality
        self.pick_output_fn = self.pick_ensemble_output
        self.rng = np.random.default_rng(seed=seed)
        self.exp_name = exp_name


    def compute_quality(
            self,
            record_set: DataRecordSet,
            expected_record_set: DataRecordSet | None = None,
            champion_record_set: DataRecordSet | None = None,
            is_filter_op: bool = False,
            is_convert_op: bool = False,
            field_to_metric_fn: dict[str, str | Callable] | None = None,
        ) -> DataRecordSet:
        """
        Compute the quality for the given `record_set` by comparing it to the `expected_record_set`.

        Update the record_set by assigning the quality to each entry in its record_op_stats and
        returning the updated record_set.
        """
        # compute whether we can only use the champion
        only_using_champion = expected_record_set is None

        # if this operation is a failed convert
        if is_convert_op and len(record_set) == 0:
            record_set.record_op_stats[0].quality = 0.0

        # if this operation is a filter:
        # - we assign a quality of 1.0 if the record is in the expected outputs and it passes this filter
        # - we assign a quality of 0.0 if the record is in the expected outputs and it does NOT pass this filter
        # - we assign a quality relative to the champion / ensemble output if the record is not in the expected outputs
        # we cannot know for certain what the correct behavior is a given filter on a record which is not in the output
        # (unless it is the only filter in the plan), thus we only evaluate the filter based on its performance on
        # records which are in the output
        elif is_filter_op:
            # NOTE:
            # - we know that record_set.record_op_stats will contain a single entry for a filter op
            # - if we are using the champion, then champion_record_set will also contain a single entry for a filter op
            record_op_stats = record_set.record_op_stats[0]
            if only_using_champion:
                champion_record = champion_record_set[0]
                record_op_stats.quality = int(record_op_stats.passed_operator == champion_record.passed_operator)

            # - if we are using validation data, we may have multiple expected records in the expected_record_set for this source_id,
            #   thus, if we can identify an exact match, we can use that to evaluate the filter's quality
            # - if we are using validation data but we *cannot* find an exact match, then we will once again use the champion record set
            else:
                # compute number of matches between this record's computed fields and this expected record's outputs
                found_match_in_output = False
                for expected_record in expected_record_set:
                    all_correct = True
                    for field, value in record_op_stats.record_state.items():
                        if value != getattr(expected_record, field):
                            all_correct = False
                            break

                    if all_correct:
                        found_match_in_output = True
                        break

                if found_match_in_output:
                    record_op_stats.quality = int(record_op_stats.passed_operator == expected_record.passed_operator)
                else:
                    champion_record = champion_record_set[0]
                    record_op_stats.quality = int(record_op_stats.passed_operator == champion_record.passed_operator)

        # if this is a successful convert operation
        else:
            # NOTE: the following computation assumes we do not project out computed values
            #       (and that the validation examples provide all computed fields); even if
            #       a user program does add projection, we can ignore the projection on the
            #       validation dataset and use the champion model (as opposed to the validation
            #       output) for scoring fields which have their values projected out

            # set the expected_record_set to be the champion_record_set if we do not have validation data
            expected_record_set = champion_record_set if only_using_champion else expected_record_set

            # GREEDY ALGORITHM
            # for each record in the expected output, we look for the computed record which maximizes the quality metric;
            # once we've identified that computed record we remove it from consideration for the next expected output
            for expected_record in expected_record_set:
                best_quality, best_record_op_stats = 0.0, None
                for record_op_stats in record_set.record_op_stats:
                    # if we already assigned this record a quality, skip it
                    if record_op_stats.quality is not None:
                        continue

                    # compute number of matches between this record's computed fields and this expected record's outputs
                    total_quality = 0
                    for field in record_op_stats.generated_fields:
                        computed_value = record_op_stats.record_state.get(field, None)
                        expected_value = getattr(expected_record, field)

                        # get the metric function for this field
                        metric_fn = (
                            field_to_metric_fn[field]
                            if field_to_metric_fn is not None and field in field_to_metric_fn
                            else "exact"
                        )

                        # compute exact match
                        if metric_fn == "exact":
                            total_quality += int(computed_value == expected_value)

                        # compute UDF metric
                        elif callable(metric_fn):
                            total_quality += metric_fn(computed_value, expected_value)

                        # otherwise, throw an exception
                        else:
                            raise Exception(f"Unrecognized metric_fn: {metric_fn}")

                    # compute recall and update best seen so far
                    quality = total_quality / len(record_op_stats.generated_fields)
                    if quality > best_quality:
                        best_quality = quality
                        best_record_op_stats = record_op_stats

                # set best_quality as quality for the best_record_op_stats
                if best_record_op_stats is not None:
                    best_record_op_stats.quality = best_quality

        # for any records which did not receive a quality, set it to 0.0 as these are unexpected extras
        for record_op_stats in record_set.record_op_stats:
            if record_op_stats.quality is None:
                record_op_stats.quality = 0.0

        return record_set


    def score_quality(
            self,
            operator_sets: list[list[PhysicalOperator]],
            execution_data: dict[str, dict[str, list[DataRecordSet]]],
            champion_outputs: dict[str, dict[str, DataRecordSet]],
            expected_outputs: dict[str, DataRecordSet] | None = None,
            field_to_metric_fn: dict[str, str | Callable] | None = None,
        ) -> list[RecordOpStats]:
        """
        NOTE: This approach to cost modeling does not work directly for aggregation queries;
              for these queries, we would ask the user to provide validation data for the step immediately
              before a final aggregation

        NOTE: This function currently assumes that one-to-many converts do NOT create duplicate outputs.
        This assumption would break if, for example, we extracted the breed of every dog in an image.
        If there were two golden retrievers and a bernoodle in an image and we extracted:

            {"image": "file1.png", "breed": "Golden Retriever"}
            {"image": "file1.png", "breed": "Golden Retriever"}
            {"image": "file1.png", "breed": "Bernedoodle"}
        
        This function would currently give perfect accuracy to the following output:

            {"image": "file1.png", "breed": "Golden Retriever"}
            {"image": "file1.png", "breed": "Bernedoodle"}

        Even though it is missing one of the golden retrievers.
        """
        # extract information about the logical operation performed at this stage of the sentinel plan;
        # NOTE: we can infer these fields from context clues, but in the long-term we should have a more
        #       principled way of getting these directly from attributes either stored in the sentinel_plan
        #       or in the PhysicalOperator
        op_set = operator_sets[-1]
        physical_op = op_set[0]
        is_source_op = isinstance(physical_op, (MarshalAndScanDataOp, CacheScanDataOp))
        is_filter_op = isinstance(physical_op, FilterOp)
        is_convert_op = isinstance(physical_op, ConvertOp)
        is_perfect_quality_op = (
            not isinstance(physical_op, LLMConvert)
            and not isinstance(physical_op, LLMFilter)
            and not isinstance(physical_op, RetrieveOp)
        )
        logical_op_id = physical_op.logical_op_id

        # if this logical_op_id is not in the execution_data (because all upstream records were filtered), return
        if logical_op_id not in execution_data:
            return execution_data

        # pull out the execution data from this operator; place the upstream execution data in a new list
        this_op_execution_data = execution_data[logical_op_id]

        # compute quality of each output computed by this operator
        for source_id, record_sets in this_op_execution_data.items():
            # NOTE
            # source_id is a particular input, for which we may have computed multiple output record_sets;
            # each of these record_sets may contain more than one record (b/c one-to-many) and we have one
            # record_set per operator in the op_set

            # if this operation does not involve an LLM, every record_op_stats object gets perfect quality
            if is_perfect_quality_op:
                for record_set in record_sets:
                    for record_op_stats in record_set.record_op_stats:
                        record_op_stats.quality = 1.0
                continue

            # get the expected output for this source_id if we have one
            expected_record_set = (
                expected_outputs[source_id]
                if expected_outputs is not None and source_id in expected_outputs
                else None
            )

            # extract champion output for this record set
            champion_record_set = champion_outputs[logical_op_id][source_id]

            # for each record_set produced by an operation, compute its quality
            for record_set in record_sets:
                record_set = self.compute_quality(record_set, expected_record_set, champion_record_set, is_filter_op, is_convert_op, field_to_metric_fn)

        # if this operator is a source op (i.e. has no input logical operator), return the execution data
        if is_source_op:
            return execution_data

        # recursively call the function on the next logical operator until you reach a scan
        execution_data = self.score_quality(operator_sets[:-1], execution_data, champion_outputs, expected_outputs, field_to_metric_fn)

        # return the quality annotated record op stats
        return execution_data


    def pick_ensemble_output(self, op_set_record_sets: list[tuple[DataRecordSet, PhysicalOperator]]) -> DataRecordSet:
        # if there's only one operator in the set, we return its record_set
        if len(op_set_record_sets) == 1:
            record_set, _ = op_set_record_sets[0]
            return record_set

        # NOTE: I don't like that this assumes the models are consistent in
        #       how they order their record outputs for one-to-many converts;
        #       eventually we can try out more robust schemes to account for
        #       differences in ordering
        # aggregate records at each index in the response
        idx_to_records = {}
        for record_set, _ in op_set_record_sets:
            for idx, record in enumerate(record_set):
                if idx not in idx_to_records:
                    idx_to_records[idx] = [record]
                else:
                    idx_to_records[idx].append(record)

        # compute most common answer at each index
        out_records = []
        for idx in range(len(idx_to_records)):
            records = idx_to_records[idx]
            most_common_record = max(set(records), key=records.count)
            out_records.append(most_common_record)

        # create and return final DataRecordSet
        return DataRecordSet(out_records, [])


    @staticmethod
    def execute_op_wrapper(operator: PhysicalOperator, op_input: DataRecord | list[DataRecord]) -> tuple[DataRecordSet, PhysicalOperator]:
        """
        Wrapper function around operator execution which also and returns the operator.
        This is useful in the parallel setting(s) where operators are executed by a worker pool,
        and it is convenient to return the op_id along with the computation result.
        """
        record_set = operator(op_input)

        return record_set, operator, op_input


    def execute_op_set(self, candidates, op_set):
        # TODO: post-submission we will need to modify this to:
        # - submit all candidates for aggregate operators
        # - handle limits
        # create thread pool w/max workers and run futures over worker pool
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            # create futures
            futures = []
            for candidate in candidates:
                for operator in op_set:
                    future = executor.submit(RandomSamplingSentinelExecutionEngine.execute_op_wrapper, operator, candidate)
                    futures.append(future)

            # compute output record_set for each (operator, candidate) pair
            output_record_sets = []
            while len(futures) > 0:
                # get the set of futures that have (and have not) finished in the last PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS
                done_futures, not_done_futures = wait(futures, timeout=PARALLEL_EXECUTION_SLEEP_INTERVAL_SECS)

                # cast not_done_futures from a set to a list so we can append to it
                not_done_futures = list(not_done_futures)

                # process finished futures
                for future in done_futures:
                    # get the result and add it to the output records set
                    record_set, operator, candidate = future.result()
                    output_record_sets.append((record_set, operator, candidate))

                # update list of futures
                futures = not_done_futures

            # compute mapping from source_id to record sets for all operators and for champion operator
            all_record_sets, champion_record_sets = {}, {}
            for candidate in candidates:
                candidate_output_record_sets = []
                for record_set, operator, candidate_ in output_record_sets:
                    if candidate == candidate_:
                        candidate_output_record_sets.append((record_set, operator))

                # select the champion (i.e. best) record_set from all the record sets computed for this operator
                champion_record_set = self.pick_output_fn(candidate_output_record_sets)

                # get the source_id associated with this input record
                source_id = candidate.source_id

                # add champion record_set to mapping from source_id --> champion record_set
                champion_record_sets[source_id] = champion_record_set

                # add all record_sets computed for this source_id to mapping from source_id --> record_sets
                all_record_sets[source_id] = [tup[0] for tup in candidate_output_record_sets]

        return all_record_sets, champion_record_sets


    def execute_sentinel_plan(self, plan: SentinelPlan, expected_outputs: dict[str, DataRecordSet], policy: Policy):
        """
        """
        if self.verbose:
            print("----------------------")
            print(f"PLAN[{plan.plan_id}] (sentinel):")
            print(plan)
            print("---")

        plan_start_time = time.time()

        # initialize plan stats and operator stats
        plan_stats = PlanStats(plan_id=plan.plan_id, plan_str=str(plan))
        for logical_op_id, logical_op_name, op_set in plan:
            op_set_details = {
                op.op_name(): {k: str(v) for k, v in op.get_id_params().items()}
                for op in op_set
            }
            plan_stats.operator_stats[logical_op_id] = OperatorStats(
                op_id=logical_op_id,
                op_name=logical_op_name,
                op_details=op_set_details,
            )

        # sample validation records
        total_num_samples = self.datasource.get_val_length()
        sample_indices = np.arange(total_num_samples)
        if self.sample_start_idx is not None:
            assert self.sample_end_idx is not None
            sample_indices = sample_indices[self.sample_start_idx:self.sample_end_idx]
        elif not self.sample_all_records:
            self.rng.shuffle(sample_indices)
            j = min(self.j, len(sample_indices))
            sample_indices = sample_indices[:j]

        # initialize output variables
        all_outputs, champion_outputs = {}, {}

        # create initial set of candidates for source scan operator
        candidates = []
        for sample_idx in sample_indices:
            candidate = DataRecord(schema=SourceRecord, source_id=sample_idx)
            candidate.idx = sample_idx
            candidate.get_item_fn = partial(self.datasource.get_item, val=True)
            candidates.append(candidate)

        # NOTE: because we need to dynamically create sample matrices for each operator,
        #       sentinel execution must be executed one operator at a time (i.e. sequentially)
        # execute operator sets in sequence
        for op_idx, (logical_op_id, _, op_set) in enumerate(plan):
            prev_logical_op_id = plan.logical_op_ids[op_idx - 1] if op_idx > 0 else None
            next_logical_op_id = plan.logical_op_ids[op_idx + 1] if op_idx + 1 < len(plan) else None

            # sample k optimizations
            k = min(self.k, len(op_set)) if not self.sample_all_ops else len(op_set)
            sampled_ops = self.rng.choice(op_set, size=k, replace=False)

            # run sampled operators on sampled candidates
            source_id_to_record_sets, source_id_to_champion_record_set = self.execute_op_set(candidates, sampled_ops)

            # update all_outputs and champion_outputs dictionary
            if logical_op_id not in all_outputs:
                all_outputs[logical_op_id] = source_id_to_record_sets
                champion_outputs[logical_op_id] = source_id_to_champion_record_set
            else:
                for source_id, record_sets in source_id_to_record_sets.items():
                    if source_id not in all_outputs[logical_op_id]:
                        all_outputs[logical_op_id][source_id] = record_sets
                        champion_outputs[logical_op_id][source_id] = source_id_to_champion_record_set[source_id]
                    else:
                        all_outputs[logical_op_id][source_id].extend(record_sets)
                        champion_outputs[logical_op_id][source_id].extend(source_id_to_champion_record_set[source_id])

            # flatten lists of records and record_op_stats
            all_records, all_record_op_stats = [], []
            for _, record_sets in source_id_to_record_sets.items():
                for record_set in record_sets:
                    all_records.extend(record_set.data_records)
                    all_record_op_stats.extend(record_set.record_op_stats)

            # update plan stats
            plan_stats.operator_stats[logical_op_id].add_record_op_stats(
                all_record_op_stats,
                source_op_id=prev_logical_op_id,
                plan_id=plan.plan_id,
            )

            # add records (which are not filtered) to the cache, if allowed
            if not self.nocache:
                for record in all_records:
                    if getattr(record, "passed_operator", True):
                        self.datadir.append_cache(logical_op_id, record)

            # update candidates for next operator; we use champion outputs as input
            candidates = []
            if next_logical_op_id is not None:
                for _, record_set in source_id_to_champion_record_set.items():
                    for record in record_set:
                        if isinstance(op_set[0], FilterOp) and not record.passed_operator:
                            continue
                        candidates.append(record)

            # if we've filtered out all records, terminate early
            if next_logical_op_id is not None and candidates == []:
                break

        # compute quality for each operator
        field_to_metric_fn = self.datasource.get_field_to_metric_fn()
        all_outputs = self.score_quality(plan.operator_sets, all_outputs, champion_outputs, expected_outputs, field_to_metric_fn)

        # if caching was allowed, close the cache
        if not self.nocache:
            for logical_op_id, _, _ in plan:
                self.datadir.close_cache(logical_op_id)

        # finalize plan stats
        total_plan_time = time.time() - plan_start_time
        plan_stats.finalize(total_plan_time)

        return all_outputs, plan_stats


    def generate_sample_observations(self, sentinel_plan: SentinelPlan, policy: Policy):
        """
        This function is responsible for generating sample observation data which can be
        consumed by the CostModel.

        To accomplish this, we construct a special sentinel plan using the Optimizer which is
        capable of executing any valid physical implementation of a Filter or Convert operator
        on each record.
        """
        # if we're using validation data, get the set of expected output records
        expected_outputs = {}
        for idx in range(self.datasource.get_val_length()):
            data_records = self.datasource.get_item(idx, val=True, include_label=True)
            if not isinstance(data_records, list):
                data_records = [data_records]
            record_set = DataRecordSet(data_records, None)
            expected_outputs[record_set.source_id] = record_set

        # run sentinel plan
        execution_data, plan_stats = self.execute_sentinel_plan(sentinel_plan, expected_outputs, policy)

        return execution_data, plan_stats


    def create_sentinel_plan(self, dataset: Set, policy: Policy) -> SentinelPlan:
        """
        Generates and returns a SentinelPlan for the given dataset.
        """
        # TODO: explicitly pull up filters; for SIGMOD we can explicitly write plans w/filters pulled up
        # initialize the optimizer
        optimizer = Optimizer(
            policy=policy,
            cost_model=CostModel(),
            no_cache=True,
            verbose=self.verbose,
            available_models=self.available_models,
            allow_bonded_query=self.allow_bonded_query,
            allow_conventional_query=self.allow_conventional_query,
            allow_code_synth=self.allow_code_synth,
            allow_token_reduction=self.allow_token_reduction,
            allow_rag_reduction=self.allow_rag_reduction,
            allow_mixtures=self.allow_mixtures,
            optimization_strategy=OptimizationStrategy.SENTINEL,
        )

        # use optimizer to generate sentinel plans
        sentinel_plans = optimizer.optimize(dataset, policy)
        sentinel_plan = sentinel_plans[0]

        return sentinel_plan


    def execute(self, dataset: Set, policy: Policy):
        execution_start_time = time.time()

        # for now, enforce that we are using validation data; we can relax this after paper submission
        if not self.using_validation_data:
            raise Exception("Make sure you are using ValidationDataSource with MABSentinelExecutionEngine")

        # if nocache is True, make sure we do not re-use codegen examples
        if self.nocache:
            self.clear_cached_examples()

        # create sentinel plan
        sentinel_plan = self.create_sentinel_plan(dataset, policy)

        # generate sample execution data
        all_execution_data, plan_stats = self.generate_sample_observations(sentinel_plan, policy)

        # put sentinel plan execution stats into list and prepare list of output records
        all_plan_stats = [plan_stats]
        all_records = []

        # construct the CostModel with any sample execution data we've gathered
        cost_model = SampleBasedCostModel(sentinel_plan, all_execution_data, self.verbose, self.exp_name)

        # (re-)initialize the optimizer
        optimizer = Optimizer(
            policy=policy,
            cost_model=cost_model,
            no_cache=self.nocache,
            verbose=self.verbose,
            available_models=self.available_models,
            allow_bonded_query=self.allow_bonded_query,
            allow_conventional_query=self.allow_conventional_query,
            allow_code_synth=self.allow_code_synth,
            allow_token_reduction=self.allow_token_reduction,
            allow_rag_reduction=self.allow_rag_reduction,
            allow_mixtures=self.allow_mixtures,
            optimization_strategy=self.optimization_strategy,
            use_final_op_quality=self.use_final_op_quality,
        )
        total_optimization_time = time.time() - execution_start_time

        # execute plan(s) according to the optimization strategy
        if self.optimization_strategy == OptimizationStrategy.CONFIDENCE_INTERVAL:
            records, plan_stats = self.execute_confidence_interval_strategy(dataset, policy, optimizer)
            all_records.extend(records)
            all_plan_stats.extend(plan_stats)

        else:
            records, plan_stats = self.execute_strategy(dataset, policy, optimizer)
            all_records.extend(records)
            all_plan_stats.extend(plan_stats)

        # aggregate plan stats
        aggregate_plan_stats = self.aggregate_plan_stats(all_plan_stats)

        # add sentinel records and plan stats (if captured) to plan execution data
        execution_stats = ExecutionStats(
            execution_id=self.execution_id(),
            plan_stats=aggregate_plan_stats,
            total_optimization_time=total_optimization_time,
            total_execution_time=time.time() - execution_start_time,
            total_execution_cost=sum(list(map(lambda plan_stats: plan_stats.total_plan_cost, aggregate_plan_stats.values()))),
            plan_strs={plan_id: plan_stats.plan_str for plan_id, plan_stats in aggregate_plan_stats.items()},
        )

        return all_records, execution_stats


class RandomSamplingSequentialSingleThreadSentinelExecution(RandomSamplingSentinelExecutionEngine, SequentialSingleThreadPlanExecutor):
    """
    This class performs sentinel execution while executing plans in a sequential, single-threaded fashion.
    """
    def __init__(self, *args, **kwargs):
        RandomSamplingSentinelExecutionEngine.__init__(self, *args, **kwargs)
        SequentialSingleThreadPlanExecutor.__init__(self, *args, **kwargs)


class RandomSamplingSequentialParallelSentinelExecution(RandomSamplingSentinelExecutionEngine, SequentialSingleThreadPlanExecutor):
    """
    This class performs sentinel execution while executing plans in a pipelined, parallel fashion.
    """
    def __init__(self, *args, **kwargs):
        RandomSamplingSentinelExecutionEngine.__init__(self, *args, **kwargs)
        # TODO: post-submission, change to parallel plan executor
        SequentialSingleThreadPlanExecutor.__init__(self, *args, **kwargs)



from __future__ import annotations

import math

# NOTE: the answer.mode() call(s) inside of _est_quality() throw a UserWarning when there are multiple
#       answers to a convert with the same mode. This is because pandas tries to sort the answers
#       before returning them, but since answer is a column of dicts the '<' operator fails on dicts.
#       For now, we can simply ignore the warning b/c we pick an answer at random anyways if there are
#       multiple w/the same count, but in the future we may want to cast the 'dict' --> 'str' or compute
#       the mode on a per-field basis.
import warnings
from typing import Any

import pandas as pd
import scipy.stats as stats

from palimpzest.constants import MODEL_CARDS, GPT_4o_MODEL_CARD, Model
from palimpzest.core.data.dataclasses import OperatorCostEstimates, PlanCost, RecordOpStats
from palimpzest.core.elements.records import DataRecordSet
from palimpzest.datamanager.datamanager import DataDirectory
from palimpzest.query.operators.aggregate import ApplyGroupByOp, AverageAggregateOp, CountAggregateOp
from palimpzest.query.operators.code_synthesis_convert import CodeSynthesisConvert
from palimpzest.query.operators.convert import LLMConvert
from palimpzest.query.operators.datasource import CacheScanDataOp, DataSourcePhysicalOp, MarshalAndScanDataOp
from palimpzest.query.operators.filter import LLMFilter, NonLLMFilter
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.rag_convert import RAGConvert
from palimpzest.query.operators.token_reduction_convert import TokenReducedConvert
from palimpzest.query.optimizer.plan import SentinelPlan
from palimpzest.utils.model_helpers import get_champion_model_name, get_models

warnings.simplefilter(action='ignore', category=UserWarning)

class BaseCostModel:
    """
    This base class contains the interface/abstraction that every CostModel must implement
    in order to work with the Optimizer. In brief, the Optimizer expects the CostModel to
    make a prediction about the runtime, cost, and quality of a physical operator.
    """
    def __init__(self):
        """
        CostModel constructor; the arguments for individual CostModels may vary depending
        on the assumptions they make about the prevalance of historical execution data
        and online vs. batch execution settings.
        """
        pass

    def get_costed_phys_op_ids(self) -> set[str]:
        """
        Return the set of physical op ids which the cost model has cost estimates for.
        """
        raise NotImplementedError("Calling get_costed_phys_op_ids from abstract method")

    def __call__(self, operator: PhysicalOperator) -> PlanCost:
        """
        The interface exposed by the CostModel to the Optimizer. Subclasses may require
        additional arguments in order to make their predictions.
        """
        raise NotImplementedError("Calling __call__ from abstract method")


class SampleBasedCostModel:
    """
    """
    def __init__(
        self,
        sentinel_plan: SentinelPlan,
        execution_data: dict[str, dict[str, list[DataRecordSet]]],
        verbose: bool = False,
        exp_name: str | None = None,
    ):
        # store sentinel plan
        self.sentinel_plan = sentinel_plan

        # store verbose argument
        self.verbose = verbose

        # store experiment name if one is provided
        self.exp_name = exp_name

        # construct cost, time, quality, and selectivity matrices for each operator set;
        self.operator_to_stats = self.compute_operator_stats(execution_data)

        # compute set of costed physical op ids from operator_to_stats
        self.costed_phys_op_ids = set([
            phys_op_id
            for _, phys_op_id_to_stats in self.operator_to_stats.items()
            for phys_op_id, _ in phys_op_id_to_stats.items()
        ])

        # reference to data directory
        self.datadir = DataDirectory()

        # import pdb; pdb.set_trace()

    def get_costed_phys_op_ids(self):
        return self.costed_phys_op_ids


    def compute_operator_stats(
            self,
            execution_data: dict[str, dict[str, list[DataRecordSet]]],
        ):
        # flatten the nested dictionary of execution data and pull out fields relevant to cost estimation
        execution_record_op_stats = []
        for idx, (logical_op_id, _, _) in enumerate(self.sentinel_plan):
            # initialize variables
            upstream_logical_op_id = self.sentinel_plan.logical_op_ids[idx - 1] if idx > 0 else None

            # filter for the execution data from this operator set
            op_set_execution_data = execution_data[logical_op_id]

            # flatten the execution data into a list of RecordOpStats
            op_set_execution_data = [
                record_op_stats
                for _, record_sets in op_set_execution_data.items()
                for record_set in record_sets
                for record_op_stats in record_set.record_op_stats
            ]

            # add entries from execution data into matrices
            for record_op_stats in op_set_execution_data:
                record_op_stats_dict = {
                    "logical_op_id": logical_op_id,
                    "physical_op_id": record_op_stats.op_id,
                    "upstream_logical_op_id": upstream_logical_op_id,
                    "record_id": record_op_stats.record_id,
                    "record_parent_id": record_op_stats.record_parent_id,
                    "cost_per_record": record_op_stats.cost_per_record,
                    "time_per_record": record_op_stats.time_per_record,
                    "quality": record_op_stats.quality,
                    "passed_operator": record_op_stats.passed_operator,
                    "source_id": record_op_stats.record_source_id,  # TODO: remove
                    "op_details": record_op_stats.op_details,       # TODO: remove
                    "answer": record_op_stats.answer,               # TODO: remove
                }
                execution_record_op_stats.append(record_op_stats_dict)

        # convert flattened execution data into dataframe
        operator_stats_df = pd.DataFrame(execution_record_op_stats)

        # for each physical_op_id, compute its average cost_per_record, time_per_record, selectivity, and quality
        operator_to_stats = {}
        for logical_op_id, logical_op_df in operator_stats_df.groupby("logical_op_id"):
            operator_to_stats[logical_op_id] = {}

            # get the logical_op_id of the upstream operator
            upstream_logical_op_ids = logical_op_df.upstream_logical_op_id.unique()
            assert len(upstream_logical_op_ids) == 1, "More than one upstream logical_op_id"
            upstream_logical_op_id = upstream_logical_op_ids[0]

            for physical_op_id, physical_op_df in logical_op_df.groupby("physical_op_id"):
                # find set of parent records for this operator
                num_upstream_records = len(physical_op_df.record_parent_id.unique())

                # compute selectivity 
                selectivity = (
                    1.0 if upstream_logical_op_id is None else physical_op_df.passed_operator.sum() / num_upstream_records
                )

                operator_to_stats[logical_op_id][physical_op_id] = {
                    "cost": physical_op_df.cost_per_record.mean(),
                    "time": physical_op_df.time_per_record.mean(),
                    "quality": physical_op_df.quality.mean(),
                    "selectivity": selectivity,
                }

        # if this is an experiment, log the dataframe and operator_to_stats dictionary
        if self.exp_name is not None:
            operator_stats_df.to_csv(f"opt-profiling-data/{self.exp_name}-operator-stats.csv", index=False)

        return operator_to_stats


    def __call__(self, operator: PhysicalOperator, source_op_estimates: OperatorCostEstimates | None = None) -> PlanCost:
        # NOTE: some physical operators may not have any sample execution data in this cost model;
        #       these physical operators are filtered out of the Optimizer, thus we can assume that
        #       we will have execution data for each operator passed into __call__; nevertheless, we
        #       still perform a sanity check
        # look up physical and logical op ids associated with this physical operator
        phys_op_id = operator.get_op_id()
        logical_op_id = operator.logical_op_id
        assert self.operator_to_stats.get(logical_op_id).get(phys_op_id) is not None, f"No execution data for {str(operator)}"

        # look up stats for this operation
        est_cost_per_record = self.operator_to_stats[logical_op_id][phys_op_id]["cost"]
        est_time_per_record = self.operator_to_stats[logical_op_id][phys_op_id]["time"]
        est_quality = self.operator_to_stats[logical_op_id][phys_op_id]["quality"]
        est_selectivity = self.operator_to_stats[logical_op_id][phys_op_id]["selectivity"]

        # create source_op_estimates for datasources if they are not provided
        if isinstance(operator, DataSourcePhysicalOp):
            # get handle to DataSource and pre-compute its size (number of records)
            datasource = self.datadir.get_registered_dataset(operator.dataset_id)
            datasource_len = len(datasource)

            source_op_estimates = OperatorCostEstimates(
                cardinality=datasource_len,
                time_per_record=0.0,
                cost_per_record=0.0,
                quality=1.0,
            )

        # generate new set of OperatorCostEstimates
        op_estimates = OperatorCostEstimates(
            cardinality=est_selectivity * source_op_estimates.cardinality,
            time_per_record=est_time_per_record,
            cost_per_record=est_cost_per_record,
            quality=est_quality,
        )

        # compute estimates for this operator
        op_time = op_estimates.time_per_record * source_op_estimates.cardinality
        op_cost = op_estimates.cost_per_record * source_op_estimates.cardinality
        op_quality = op_estimates.quality

        # construct and return op estimates
        return PlanCost(cost=op_cost, time=op_time, quality=op_quality, op_estimates=op_estimates)


class CostModel(BaseCostModel):
    """
    This class takes in a list of RecordOpStats and performs cost estimation on a given operator
    by taking the average of any sample execution that the CostModel has for that operator. If no
    such data exists, it returns a naive estimate.
    """
    def __init__(
        self,
        sample_execution_data: list[RecordOpStats] | None = None,
        available_models: list[Model] | None = None,
        confidence_level: float = 0.90,
    ) -> None:
        if sample_execution_data is None:
            sample_execution_data = []
        if available_models is None:
            available_models = []

        # construct full dataset of samples
        self.sample_execution_data_df = (
            pd.DataFrame(sample_execution_data)
            if len(sample_execution_data) > 0
            else None
        )
        # df contains a column called record_state, that sometimes contain a dict
        # we want to extract the keys from the dict and create a new column for each key

        # reference to data directory
        self.datadir = DataDirectory()

        # set available models
        self.available_models = available_models

        # set confidence level for CI estimates
        self.conf_level = confidence_level

        # compute per-operator estimates
        self.operator_estimates = self._compute_operator_estimates()

        # compute set of costed physical op ids from operator_to_stats
        self.costed_phys_op_ids = None if self.operator_estimates is None else set(self.operator_estimates.keys())

    def get_costed_phys_op_ids(self):
        return self.costed_phys_op_ids

    def _compute_ci(self, sample_mean: float, n_samples: int, std_dev: float) -> tuple[float, float]:
        """
        Compute confidence interval (for non-proportion quantities) given the sample mean, number of samples,
        and sample std. deviation at the CostModel's given confidence level. We use a t-distribution for
        computing the interval as many sample estimates in PZ may have few samples.
        """
        ci = stats.t.interval(
            confidence=self.conf_level,  # Confidence level
            df=n_samples - 1,            # Degrees of freedom
            loc=sample_mean,             # Sample mean
            scale=std_dev,               # Standard deviation estimate
        )
        return ci

    def _compute_proportion_ci(self, sample_prop: float, n_samples: int) -> tuple[float, float]:
        """
        Compute confidence interval for proportion quantities (i.e. selectivity) given the sample proportion
        and the number of samples. We use the normal distribution for computing the interval here, for reasons
        summarized by this post: https://stats.stackexchange.com/a/411727.
        """
        if sample_prop == 0.0 or sample_prop == 1.0:
            return (sample_prop, sample_prop)

        scaling_factor = math.sqrt((sample_prop * (1 - sample_prop)) / n_samples)
        lower_bound, upper_bound = stats.norm.interval(
            confidence=self.conf_level,  # Confidence level
            loc=sample_prop,             # Sample proportion
            scale=scaling_factor,        # Scaling factor
        )
        lower_bound = max(lower_bound, 0.0)
        upper_bound = max(upper_bound, 1.0)

        return (lower_bound, upper_bound)

    def _compute_mean_and_ci(self, df: pd.DataFrame, col: str, model_name: str | None = None, non_negative_lb: bool = False) -> tuple[float, float, float]:
        """
        Compute the mean and CI for the given column and dataframe. If the model_name is provided, filter
        for the subset of rows belonging to the model.
        """
        # use model-specific estimate if possible
        if model_name is not None:
            model_df = df[df.model_name == model_name]
            if not model_df.empty:
                col_mean = model_df[col].mean()
                col_lb, col_ub = self._compute_ci(
                    sample_mean=col_mean,
                    n_samples=model_df[col].notna().sum(),
                    std_dev=model_df[col].std(),
                )
                if non_negative_lb:
                    col_lb = max(col_lb, 0.0)

                return col_mean, col_lb, col_ub

        # compute aggregate
        col_mean = df[col].mean()
        col_lb, col_ub = self._compute_ci(
            sample_mean=col_mean,
            n_samples=df[col].notna().sum(),
            std_dev=df[col].std(),
        )
        if non_negative_lb:
            col_lb = max(col_lb, 0.0)

        return col_mean, col_lb, col_ub

    def _est_time_per_record(self, op_df: pd.DataFrame, model_name: str | None = None) -> tuple[float, float, float]:
        """
        Given sample cost data observations for a specific operation, compute the mean and CI
        for the time per record.
        """
        return self._compute_mean_and_ci(df=op_df, col="time_per_record", model_name=model_name, non_negative_lb=True)

    def _est_cost_per_record(self, op_df: pd.DataFrame, model_name: str | None = None) -> tuple[float, float, float]:
        """
        Given sample cost data observations for a specific operation, compute the mean and CI
        for the cost per record.
        """
        return self._compute_mean_and_ci(df=op_df, col="cost_per_record", model_name=model_name, non_negative_lb=True)

    def _est_tokens_per_record(self, op_df: pd.DataFrame, model_name: str | None = None) -> tuple[tuple[float, float, float], tuple[float, float, float]]:
        """
        Given sample cost data observations for a specific operation, compute the mean and CI
        for the total input tokens and total output tokens.
        """
        total_input_tokens_tuple = self._compute_mean_and_ci(df=op_df, col="total_input_tokens", model_name=model_name, non_negative_lb=True)
        total_output_tokens_tuple = self._compute_mean_and_ci(df=op_df, col="total_output_tokens", model_name=model_name, non_negative_lb=True)

        return total_input_tokens_tuple, total_output_tokens_tuple

    def _est_cardinality(self, op_df: pd.DataFrame, model_name: str | None = None) -> float:
        """
        Given sample cost data observations for a specific operation, compute the number of
        rows output by the operation.

        NOTE: right now, this should only be used by the ApplyGroupByOp as a way to gauge the
        number of output groups. Using this to estimate, e.g. the cardinality of a filter,
        convert, or base scan will lead to wildly inaccurate results because the absolute value
        of these cardinalities will simply be a reflection of the sample size.

        For those operations, we use the `_est_selectivity` function to estimate the operator's
        selectivity, which we can apply to an est. of the operator's input cardinality.
        """
        return op_df.shape[0] / len(op_df.plan_id.unique())

    def _est_selectivity(self, df: pd.DataFrame, op_df: pd.DataFrame, model_name: str | None = None) -> float:
        """
        Given sample cost data observations for the plan and a specific operation, compute
        the ratio of records between this operator and its source operator.
        """
        # compute whether or not this operation is a filter
        is_filter_op = "filter" in str(op_df.op_name.iloc[0]).lower()

        # use model-specific estimate if possible
        if model_name is not None:
            model_op_df = op_df[op_df.model_name == model_name]
            if not model_op_df.empty:
                num_input_records = model_op_df.shape[0]

                # get subset of records that were the source to this operator
                num_output_records = None
                if is_filter_op:
                    num_output_records = model_op_df.passed_operator.sum()
                else:
                    op_ids = model_op_df.op_id.unique().tolist()
                    plan_ids = model_op_df.plan_id.unique().tolist()
                    num_output_records = df[df.source_op_id.isin(op_ids) & df.plan_id.isin(plan_ids)].shape[0]

                # estimate the selectivity / fan-out and compute bounds
                est_selectivity = num_output_records / num_input_records
                if is_filter_op:
                    est_selectivity_lb, est_selectivity_ub = self._compute_proportion_ci(est_selectivity, n_samples=num_input_records)

                # for now, if we are doing a convert operation w/fan-out then the assumptions of _compute_proportion_ci
                # do not hold; until we have a better method for estimating bounds, just set them to the estimate
                else:
                    est_selectivity_lb = est_selectivity
                    est_selectivity_ub = est_selectivity

                return est_selectivity, est_selectivity_lb, est_selectivity_ub

        # otherwise average selectivity across all ops
        num_input_records = op_df.shape[0]

        # get subset of records that were the source to this operator
        num_output_records = None
        if is_filter_op:
            num_output_records = op_df.passed_operator.sum()
        else:
            op_ids = op_df.op_id.unique().tolist()
            num_output_records = df[df.source_op_id.isin(op_ids)].shape[0]

        # estimate the selectivity / fan-out and compute bounds
        est_selectivity = num_output_records / num_input_records
        if is_filter_op:
            est_selectivity_lb, est_selectivity_ub = self._compute_proportion_ci(est_selectivity, n_samples=num_input_records)

        # for now, if we are doing a convert operation w/fan-out then the assumptions of _compute_proportion_ci
        # do not hold; until we have a better method for estimating bounds, just set them to the estimate
        else:
            est_selectivity_lb = est_selectivity
            est_selectivity_ub = est_selectivity

        return est_selectivity, est_selectivity_lb, est_selectivity_ub

    def _compute_quality(self, row):
        # compute accuracy for filter
        if "filter" in row["op_name"].lower():
            row["correct"] = int(row["answer"] == row["accepted_answer"])
            row["num_answers"] = 1
            return row

        # otherwise, compute recall on a per-key basis
        try:
            # we'll measure recall on accepted_answer, as extraneous info is often not an issue
            answer = row["answer"]
            accepted_answer = row["accepted_answer"]
            correct = 0
            for key, value in accepted_answer.items():
                if key in answer and answer[key] == value:
                    correct += 1
            
            row["correct"] = correct
            row["num_answers"] = len(accepted_answer.keys())
            return row

        except Exception as e:
            print(f"WARNING: error decoding answer or accepted_answer: {str(e)}")
            row["correct"] = 0
            row["num_answers"] = 1
            return row

    def _est_quality(self, op_df: pd.DataFrame, model_name: str | None = None) -> float:
        """
        Given sample cost data observations for a specific operation, compute the an estimate
        of the quality of its outputs by using GPT-4 as a champion model.
        """
        # get unique set of records
        record_ids = op_df.record_id.unique()

        # get champion model name
        vision = ("image_operation" in op_df.columns and op_df.image_operation.any())
        champion_model_name = get_champion_model_name(self.available_models, vision)

        # compute champion's answer (per-record) across all models; fall-back to most common answer if champion is not present
        record_id_to_answer = {}
        for record_id in record_ids:
            record_df = op_df[op_df.record_id == record_id]
            champion_most_common_answer = record_df[
                record_df.model_name == champion_model_name
            ].answer.mode()
            all_models_most_common_answer = record_df.answer.mode()

            if not champion_most_common_answer.empty:
                record_id_to_answer[record_id] = champion_most_common_answer.iloc[0]
            elif not all_models_most_common_answer.empty:
                record_id_to_answer[record_id] = all_models_most_common_answer.iloc[0]
            else:
                record_id_to_answer[record_id] = None

        # compute accepted answers and clean all answers
        pd.options.mode.chained_assignment = None  # turn off copy warnings
        op_df.loc[:, "accepted_answer"] = op_df.record_id.apply(lambda id: record_id_to_answer[id])
        op_df = op_df.apply(lambda row: self._compute_quality(row), axis=1)

        # get subset of observations for model_name and estimate quality w/fraction of answers that match accepted answer
        model_df = (
            op_df[op_df.model_name == model_name]
            if model_name is not None
            else op_df[op_df.model_name.isna()]
        )

        # compute quality as the fraction of answers which are correct (recall on expected output)
        num_correct = model_df.correct.sum() if not model_df.empty else op_df.correct.sum()
        total_answers = model_df.num_answers.sum() if not model_df.empty else op_df.num_answers.sum()
        est_quality = num_correct / total_answers

        # compute CI on the proportion of correct answers
        est_quality_lb, est_quality_ub = self._compute_proportion_ci(est_quality, n_samples=total_answers)

        return est_quality, est_quality_lb, est_quality_ub

    def _compute_operator_estimates(self) -> dict[str, Any] | None:
        """
        Compute per-operator estimates of runtime, cost, and quality.
        """
        # if we don't have sample execution data, we cannot compute per-operator estimates
        if self.sample_execution_data_df is None:
            return None

        # get the set of operator ids for which we have sample data
        op_ids = self.sample_execution_data_df.op_id.unique()

        # compute estimates of runtime, cost, and quality (and intermediates like cardinality) for every operator
        operator_estimates = {}
        for op_id in op_ids:
            # filter for subset of sample execution data related to this operation
            op_df = self.sample_execution_data_df[
                self.sample_execution_data_df.op_id == op_id
            ]

            # skip computing an estimate if we didn't capture any sampling data for this operator
            # (this can happen if/when upstream filter operation(s) filter out all records) 
            if op_df.empty:
                continue

            # initialize estimates
            estimates = {}

            # get the op_name for this operation
            model_name = op_df.model_name.iloc[0] if op_df.model_name.iloc[0] is not None else None
            op_name = str(op_df.op_name.iloc[0])
            if model_name is not None:
                # compute estimates per-model, and add None which forces computation of avg. across all models
                model_names = [m.value for m in get_models(include_vision=True)] + [None]
                # model_names = op_df.model_name.unique().tolist()
                estimates = {model_name: None for model_name in model_names}
                for model_name in model_names:
                    time_per_record, time_per_record_lb, time_per_record_ub = self._est_time_per_record(op_df, model_name=model_name)
                    cost_per_record, cost_per_record_lb, cost_per_record_ub = self._est_cost_per_record(op_df, model_name=model_name)
                    input_tokens_tup, output_tokens_tup = self._est_tokens_per_record(op_df, model_name=model_name)
                    selectivity, selectivity_lb, selectivity_ub = self._est_selectivity(self.sample_execution_data_df, op_df, model_name=model_name)
                    quality, quality_lb, quality_ub = self._est_quality(op_df, model_name=model_name)
                    
                    model_estimates = {
                        "time_per_record": time_per_record,
                        "time_per_record_lower_bound": time_per_record_lb,
                        "time_per_record_upper_bound": time_per_record_ub,
                        "cost_per_record": cost_per_record,
                        "cost_per_record_lower_bound": cost_per_record_lb,
                        "cost_per_record_upper_bound": cost_per_record_ub,
                        "total_input_tokens": input_tokens_tup[0],
                        "total_input_tokens_lower_bound": input_tokens_tup[1],
                        "total_input_tokens_upper_bound": input_tokens_tup[2],
                        "total_output_tokens": output_tokens_tup[0],
                        "total_output_tokens_lower_bound": output_tokens_tup[1],
                        "total_output_tokens_upper_bound": output_tokens_tup[2],
                        "selectivity": selectivity,
                        "selectivity_lower_bound": selectivity_lb,
                        "selectivity_upper_bound": selectivity_ub,
                        "quality": quality,
                        "quality_lower_bound": quality_lb,
                        "quality_upper_bound": quality_ub,
                    }
                    estimates[model_name] = model_estimates

            # TODO pre-compute lists of op_names in groups
            elif op_name in ["NonLLMFilter"]:
                time_per_record, time_per_record_lb, time_per_record_ub = self._est_time_per_record(op_df)
                selectivity, selectivity_lb, selectivity_ub = self._est_selectivity(self.sample_execution_data_df, op_df)
                estimates = {
                    "time_per_record": time_per_record,
                    "time_per_record_lower_bound": time_per_record_lb,
                    "time_per_record_upper_bound": time_per_record_ub,
                    "selectivity": selectivity,
                    "selectivity_lower_bound": selectivity_lb,
                    "selectivity_upper_bound": selectivity_ub,
                }

            elif op_name in ["MarshalAndScanDataOp", "CacheScanDataOp", "LimitScanOp", "CountAggregateOp", "AverageAggregateOp"]:
                time_per_record, time_per_record_lb, time_per_record_ub = self._est_time_per_record(op_df)
                estimates = {
                    "time_per_record": time_per_record,
                    "time_per_record_lower_bound": time_per_record_lb,
                    "time_per_record_upper_bound": time_per_record_ub,
                }

            elif op_name in ["ApplyGroupByOp"]:
                time_per_record, time_per_record_lb, time_per_record_ub = self._est_time_per_record(op_df)
                cardinality = self._est_cardinality(op_df)
                estimates = {
                    "time_per_record": time_per_record,
                    "time_per_record_lower_bound": time_per_record_lb,
                    "time_per_record_upper_bound": time_per_record_ub,
                    "cardinality": cardinality,
                }

            operator_estimates[op_id] = estimates

        return operator_estimates

    def __call__(self, operator: PhysicalOperator, source_op_estimates: OperatorCostEstimates | None = None) -> PlanCost:
        # get identifier for operation which is unique within sentinel plan but consistent across sentinels
        op_id = operator.get_op_id()

        # initialize estimates of operator metrics based on naive (but sometimes precise) logic
        if isinstance(operator, MarshalAndScanDataOp):
            # get handle to DataSource and pre-compute its size (number of records)
            datasource = self.datadir.get_registered_dataset(operator.dataset_id)
            dataset_type = self.datadir.get_registered_dataset_type(operator.dataset_id)
            datasource_len = len(datasource)
            datasource_memsize = datasource.get_size()

            source_op_estimates = OperatorCostEstimates(
                cardinality=datasource_len,
                time_per_record=0.0,
                cost_per_record=0.0,
                quality=1.0,
            )

            op_estimates = operator.naive_cost_estimates(source_op_estimates,
                                                    input_record_size_in_bytes=datasource_memsize/datasource_len,
                                                    dataset_type=dataset_type)

        elif isinstance(operator, CacheScanDataOp):
            datasource = self.datadir.get_cached_result(operator.dataset_id)
            datasource_len = len(datasource)
            datasource_memsize = datasource.get_size()

            source_op_estimates = OperatorCostEstimates(
                cardinality=datasource_len,
                time_per_record=0.0,
                cost_per_record=0.0,
                quality=1.0,
            )

            op_estimates = operator.naive_cost_estimates(source_op_estimates, input_record_size_in_bytes=datasource_memsize/datasource_len)

        else:
            op_estimates = operator.naive_cost_estimates(source_op_estimates)

        # if we have sample execution data, update naive estimates with more informed ones
        sample_op_estimates = self.operator_estimates
        if sample_op_estimates is not None and op_id in sample_op_estimates:
            if isinstance(operator, (MarshalAndScanDataOp, CacheScanDataOp)):
                op_estimates.time_per_record = sample_op_estimates[op_id]["time_per_record"]
                op_estimates.time_per_record_lower_bound = sample_op_estimates[op_id]["time_per_record_lower_bound"]
                op_estimates.time_per_record_upper_bound = sample_op_estimates[op_id]["time_per_record_upper_bound"]

            elif isinstance(operator, ApplyGroupByOp):
                # NOTE: in theory we should also treat this cardinality est. as a random variable, but in practice we will
                #       have K samples of the number of groups produced by the groupby operator, where K is the number of
                #       plans we generate sample data with. For now, we will simply use the estimate without bounds.
                #
                # NOTE: this cardinality is the only cardinality we estimate directly b/c we can observe how many groups are
                #       produced by the groupby in our sample and assume it may generalize to the full workload. To estimate
                #       actual cardinalities of operators we estimate their selectivities / fan-outs and multiply those by
                #       the input cardinality (where the initial input cardinality from the datasource is known).
                op_estimates.cardinality = sample_op_estimates[op_id]["cardinality"]
                op_estimates.cardinality_lower_bound = op_estimates.cardinality
                op_estimates.cardinality_upper_bound = op_estimates.cardinality
                op_estimates.time_per_record = sample_op_estimates[op_id]["time_per_record"]
                op_estimates.time_per_record_lower_bound = sample_op_estimates[op_id]["time_per_record_lower_bound"]
                op_estimates.time_per_record_upper_bound = sample_op_estimates[op_id]["time_per_record_upper_bound"]

            elif isinstance(operator, (CountAggregateOp, AverageAggregateOp)):  # noqa: SIM114
                op_estimates.time_per_record = sample_op_estimates[op_id]["time_per_record"]
                op_estimates.time_per_record_lower_bound = sample_op_estimates[op_id]["time_per_record_lower_bound"]
                op_estimates.time_per_record_upper_bound = sample_op_estimates[op_id]["time_per_record_upper_bound"]

            elif isinstance(operator, LimitScanOp):
                op_estimates.time_per_record = sample_op_estimates[op_id]["time_per_record"]
                op_estimates.time_per_record_lower_bound = sample_op_estimates[op_id]["time_per_record_lower_bound"]
                op_estimates.time_per_record_upper_bound = sample_op_estimates[op_id]["time_per_record_upper_bound"]

            elif isinstance(operator, NonLLMFilter):
                op_estimates.cardinality = source_op_estimates.cardinality * sample_op_estimates[op_id]["selectivity"]
                op_estimates.cardinality_lower_bound = source_op_estimates.cardinality_lower_bound * sample_op_estimates[op_id]["selectivity_lower_bound"]
                op_estimates.cardinality_upper_bound = source_op_estimates.cardinality_upper_bound * sample_op_estimates[op_id]["selectivity_upper_bound"]

                op_estimates.time_per_record = sample_op_estimates[op_id]["time_per_record"]
                op_estimates.time_per_record_lower_bound = sample_op_estimates[op_id]["time_per_record_lower_bound"]
                op_estimates.time_per_record_upper_bound = sample_op_estimates[op_id]["time_per_record_upper_bound"]

            elif isinstance(operator, LLMFilter):
                model_name = operator.model.value
                op_estimates.cardinality = source_op_estimates.cardinality * sample_op_estimates[op_id][model_name]["selectivity"]
                op_estimates.cardinality_lower_bound = source_op_estimates.cardinality_lower_bound * sample_op_estimates[op_id][model_name]["selectivity_lower_bound"]
                op_estimates.cardinality_upper_bound = source_op_estimates.cardinality_upper_bound * sample_op_estimates[op_id][model_name]["selectivity_upper_bound"]

                op_estimates.time_per_record = sample_op_estimates[op_id][model_name]["time_per_record"]
                op_estimates.time_per_record_lower_bound = sample_op_estimates[op_id][model_name]["time_per_record_lower_bound"]
                op_estimates.time_per_record_upper_bound = sample_op_estimates[op_id][model_name]["time_per_record_upper_bound"]

                op_estimates.cost_per_record = sample_op_estimates[op_id][model_name]["cost_per_record"]
                op_estimates.cost_per_record_lower_bound = sample_op_estimates[op_id][model_name]["cost_per_record_lower_bound"]
                op_estimates.cost_per_record_upper_bound = sample_op_estimates[op_id][model_name]["cost_per_record_upper_bound"]

                op_estimates.quality = sample_op_estimates[op_id][model_name]["quality"]
                op_estimates.quality_lower_bound = sample_op_estimates[op_id][model_name]["quality_lower_bound"]
                op_estimates.quality_upper_bound = sample_op_estimates[op_id][model_name]["quality_upper_bound"]

            elif isinstance(operator, LLMConvert):
                # TODO: EVEN BETTER: do similarity match (e.g. largest param intersection, more exotic techniques);
                #       another heuristic: logical_op_id-->subclass_physical_op_id-->specific_physical_op_id-->most_param_match_physical_op_id
                # TODO: instead of [op_id][model_name] --> [logical_op_id][physical_op_id]
                # NOTE: code synthesis does not have a model attribute
                model_name = operator.model.value if hasattr(operator, "model") else None
                op_estimates.cardinality = source_op_estimates.cardinality * sample_op_estimates[op_id][model_name]["selectivity"]
                op_estimates.cardinality_lower_bound = source_op_estimates.cardinality_lower_bound * sample_op_estimates[op_id][model_name]["selectivity_lower_bound"]
                op_estimates.cardinality_upper_bound = source_op_estimates.cardinality_upper_bound * sample_op_estimates[op_id][model_name]["selectivity_upper_bound"]

                op_estimates.time_per_record = sample_op_estimates[op_id][model_name]["time_per_record"]
                op_estimates.time_per_record_lower_bound = sample_op_estimates[op_id][model_name]["time_per_record_lower_bound"]
                op_estimates.time_per_record_upper_bound = sample_op_estimates[op_id][model_name]["time_per_record_upper_bound"]

                op_estimates.cost_per_record = sample_op_estimates[op_id][model_name]["cost_per_record"]
                op_estimates.cost_per_record_lower_bound = sample_op_estimates[op_id][model_name]["cost_per_record_lower_bound"]
                op_estimates.cost_per_record_upper_bound = sample_op_estimates[op_id][model_name]["cost_per_record_upper_bound"]

                op_estimates.quality = sample_op_estimates[op_id][model_name]["quality"]
                op_estimates.quality_lower_bound = sample_op_estimates[op_id][model_name]["quality_lower_bound"]
                op_estimates.quality_upper_bound = sample_op_estimates[op_id][model_name]["quality_upper_bound"]

                # NOTE: if code synth. fails, this will turn into ConventionalQuery calls to GPT-3.5,
                #       which would wildly mess up estimate of time and cost per-record
                # do code synthesis adjustment
                if isinstance(operator, CodeSynthesisConvert):
                    op_estimates.time_per_record = 1e-5
                    op_estimates.time_per_record_lower_bound = op_estimates.time_per_record
                    op_estimates.time_per_record_upper_bound = op_estimates.time_per_record
                    op_estimates.cost_per_record = 1e-4
                    op_estimates.cost_per_record_lower_bound = op_estimates.cost_per_record
                    op_estimates.cost_per_record_upper_bound = op_estimates.cost_per_record
                    op_estimates.quality = op_estimates.quality * (GPT_4o_MODEL_CARD["code"] / 100.0)
                    op_estimates.quality_lower_bound = op_estimates.quality_lower_bound * (GPT_4o_MODEL_CARD["code"] / 100.0)
                    op_estimates.quality_upper_bound = op_estimates.quality_upper_bound * (GPT_4o_MODEL_CARD["code"] / 100.0)

                # token reduction adjustment
                if isinstance(operator, TokenReducedConvert):
                    total_input_tokens = operator.token_budget * sample_op_estimates[op_id][model_name]["total_input_tokens"]
                    total_output_tokens = sample_op_estimates[op_id][model_name]["total_output_tokens"]
                    op_estimates.cost_per_record = (
                        MODEL_CARDS[model_name]["usd_per_input_token"] * total_input_tokens
                        + MODEL_CARDS[model_name]["usd_per_output_token"] * total_output_tokens
                    )
                    total_input_tokens_lb = operator.token_budget * sample_op_estimates[op_id][model_name]["total_input_tokens_lower_bound"]
                    total_output_tokens_lb = sample_op_estimates[op_id][model_name]["total_output_tokens_lower_bound"]
                    op_estimates.cost_per_record_lower_bound = (
                        MODEL_CARDS[model_name]["usd_per_input_token"] * total_input_tokens_lb
                        + MODEL_CARDS[model_name]["usd_per_output_token"] * total_output_tokens_lb
                    )
                    total_input_tokens_ub = operator.token_budget * sample_op_estimates[op_id][model_name]["total_input_tokens_upper_bound"]
                    total_output_tokens_ub = sample_op_estimates[op_id][model_name]["total_output_tokens_upper_bound"]
                    op_estimates.cost_per_record_upper_bound = (
                        MODEL_CARDS[model_name]["usd_per_input_token"] * total_input_tokens_ub
                        + MODEL_CARDS[model_name]["usd_per_output_token"] * total_output_tokens_ub
                    )

                    op_estimates.quality = op_estimates.quality * math.sqrt(math.sqrt(operator.token_budget))
                    op_estimates.quality_lower_bound = op_estimates.quality_lower_bound * math.sqrt(math.sqrt(operator.token_budget))
                    op_estimates.quality_upper_bound = op_estimates.quality_upper_bound * math.sqrt(math.sqrt(operator.token_budget))
                
                # rag convert adjustment
                if isinstance(operator, RAGConvert):
                    total_input_tokens = operator.num_chunks_per_field * operator.chunk_size
                    total_output_tokens = sample_op_estimates[op_id][model_name]["total_output_tokens"]
                    op_estimates.cost_per_record = (
                        MODEL_CARDS[model_name]["usd_per_input_token"] * total_input_tokens
                        + MODEL_CARDS[model_name]["usd_per_output_token"] * total_output_tokens
                    )
                    total_input_tokens_lb = operator.num_chunks_per_field * operator.chunk_size
                    total_output_tokens_lb = sample_op_estimates[op_id][model_name]["total_output_tokens_lower_bound"]
                    op_estimates.cost_per_record_lower_bound = (
                        MODEL_CARDS[model_name]["usd_per_input_token"] * total_input_tokens_lb
                        + MODEL_CARDS[model_name]["usd_per_output_token"] * total_output_tokens_lb
                    )
                    total_input_tokens_ub = operator.num_chunks_per_field * operator.chunk_size
                    total_output_tokens_ub = sample_op_estimates[op_id][model_name]["total_output_tokens_upper_bound"]
                    op_estimates.cost_per_record_upper_bound = (
                        MODEL_CARDS[model_name]["usd_per_input_token"] * total_input_tokens_ub
                        + MODEL_CARDS[model_name]["usd_per_output_token"] * total_output_tokens_ub
                    )

                    op_estimates.quality = op_estimates.quality * operator.naive_quality_adjustment
                    op_estimates.quality_lower_bound = op_estimates.quality_lower_bound * operator.naive_quality_adjustment
                    op_estimates.quality_upper_bound = op_estimates.quality_upper_bound * operator.naive_quality_adjustment

            else:
                raise Exception("Unknown operator")

        # compute estimates for this operator
        op_time = op_estimates.time_per_record * source_op_estimates.cardinality
        op_cost = op_estimates.cost_per_record * source_op_estimates.cardinality
        op_quality = op_estimates.quality

        # compute bounds on total time and cost estimates for this operator
        op_cost_lower_bound = op_estimates.cost_per_record_lower_bound * source_op_estimates.cardinality_lower_bound
        op_cost_upper_bound = op_estimates.cost_per_record_upper_bound * source_op_estimates.cardinality_upper_bound
        op_time_lower_bound = op_estimates.time_per_record_lower_bound * source_op_estimates.cardinality_lower_bound
        op_time_upper_bound = op_estimates.time_per_record_upper_bound * source_op_estimates.cardinality_upper_bound
        op_quality_lower_bound = op_estimates.quality_lower_bound
        op_quality_upper_bound = op_estimates.quality_upper_bound

        # create and return PlanCost object for this op's statistics
        op_plan_cost = PlanCost(
            cost=op_cost,
            time=op_time,
            quality=op_quality,
            op_estimates=op_estimates,
            cost_lower_bound=op_cost_lower_bound,
            cost_upper_bound=op_cost_upper_bound,
            time_lower_bound=op_time_lower_bound,
            time_upper_bound=op_time_upper_bound,
            quality_lower_bound=op_quality_lower_bound,
            quality_upper_bound=op_quality_upper_bound,
        )

        return op_plan_cost



from __future__ import annotations

from copy import deepcopy

from palimpzest.constants import Model, OptimizationStrategy
from palimpzest.core.data.datasources import DataSource
from palimpzest.core.lib.fields import Field
from palimpzest.datamanager.datamanager import DataDirectory
from palimpzest.policy import Policy
from palimpzest.query.operators.logical import (
    Aggregate,
    BaseScan,
    CacheScan,
    ConvertScan,
    FilteredScan,
    GroupByAggregate,
    LimitScan,
    LogicalOperator,
    Project,
    RetrieveScan,
)
from palimpzest.query.optimizer import (
    IMPLEMENTATION_RULES,
    TRANSFORMATION_RULES,
)
from palimpzest.query.optimizer.cost_model import CostModel
from palimpzest.query.optimizer.plan import PhysicalPlan, SentinelPlan
from palimpzest.query.optimizer.primitives import Group, LogicalExpression
from palimpzest.query.optimizer.rules import (
    CodeSynthesisConvertRule,
    LLMConvertBondedRule,
    LLMConvertConventionalRule,
    MixtureOfAgentsConvertRule,
    RAGConvertRule,
    TokenReducedConvertBondedRule,
    TokenReducedConvertConventionalRule,
    TokenReducedConvertRule,
)
from palimpzest.query.optimizer.tasks import (
    ApplyRule,
    ExpandGroup,
    OptimizeGroup,
    OptimizeLogicalExpression,
    OptimizePhysicalExpression,
)
from palimpzest.sets import Dataset, Set
from palimpzest.utils.model_helpers import get_champion_model, get_code_champion_model, get_conventional_fallback_model


class Optimizer:
    """
    The optimizer is responsible for searching the space of possible physical plans
    for a user's initial (logical) plan and selecting the one which is closest to
    optimizing the user's policy objective.

    This optimizer is modeled after the Cascades framework for top-down query optimization:
    - Thesis describing Cascades implementation (Chapters 1-3):
      https://15721.courses.cs.cmu.edu/spring2023/papers/17-optimizer2/xu-columbia-thesis1998.pdf

    - Andy Pavlo lecture with walkthrough example: https://www.youtube.com/watch?v=PXS49-tFLcI

    - Original Paper: https://www.cse.iitb.ac.in/infolab/Data/Courses/CS632/2015/Papers/Cascades-graefe.pdf

    Notably, this optimization framework has served as the backbone of Microsoft SQL Server, CockroachDB,
    and a few other important DBMS systems.

    NOTE: the optimizer currently assumes that field names are unique across schemas; we do try to enforce
          this by rewriting field names underneath-the-hood to be "{schema_name}.{field_name}", but this still
          does not solve a situation in which -- for example -- a user uses the pz.URL schema twice in the same
          program. In order to address that situation, we will need to augment our renaming scheme.
    """

    def __init__(
        self,
        policy: Policy,
        cost_model: CostModel,
        no_cache: bool = False,
        verbose: bool = False,
        available_models: list[Model] | None = None,
        allow_bonded_query: bool = True,
        allow_conventional_query: bool = False,
        allow_code_synth: bool = False,
        allow_token_reduction: bool = False,
        allow_rag_reduction: bool = True,
        allow_mixtures: bool = True,
        optimization_strategy: OptimizationStrategy = OptimizationStrategy.PARETO,
        use_final_op_quality: bool = False, # TODO: make this func(plan) -> final_quality
    ):
        # store the policy
        if available_models is None:
            available_models = []
        self.policy = policy

        # store the cost model
        self.cost_model = cost_model

        # store the set of physical operators for which our cost model has cost estimates
        self.costed_phys_op_ids = cost_model.get_costed_phys_op_ids()

        # mapping from each group id to its Group object
        self.groups = {}

        # mapping from each expression to its Expression object
        self.expressions = {}

        # the stack of tasks to perform during optimization
        self.tasks_stack = []

        # the lists of implementation and transformation rules that the optimizer can apply
        self.implementation_rules = IMPLEMENTATION_RULES
        self.transformation_rules = TRANSFORMATION_RULES

        # if we are doing SENTINEL / NONE optimization; remove transformation rules
        if optimization_strategy in [OptimizationStrategy.SENTINEL, OptimizationStrategy.NONE]:
            self.transformation_rules = []

        # if we are not performing optimization, set available models to be single model
        # and remove all optimizations (except for bonded queries)
        if optimization_strategy == OptimizationStrategy.NONE:
            self.allow_bonded_query = True
            self.allow_conventional_query = False
            self.allow_code_synth = False
            self.allow_token_reduction = False
            self.allow_rag_reduction = False
            self.allow_mixtures = False
            self.available_models = [available_models[0]]

        # store optimization hyperparameters
        self.no_cache = no_cache
        self.verbose = verbose
        self.available_models = available_models
        self.allow_bonded_query = allow_bonded_query
        self.allow_conventional_query = allow_conventional_query
        self.allow_code_synth = allow_code_synth
        self.allow_token_reduction = allow_token_reduction
        self.allow_rag_reduction = allow_rag_reduction
        self.allow_mixtures = allow_mixtures
        self.optimization_strategy = optimization_strategy
        self.use_final_op_quality = use_final_op_quality

        # prune implementation rules based on boolean flags
        if not self.allow_bonded_query:
            self.implementation_rules = [
                rule
                for rule in self.implementation_rules
                if rule not in [LLMConvertBondedRule, TokenReducedConvertBondedRule]
            ]

        if not self.allow_conventional_query:
            self.implementation_rules = [
                rule
                for rule in self.implementation_rules
                if rule not in [LLMConvertConventionalRule, TokenReducedConvertConventionalRule]
            ]

        if not self.allow_code_synth:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, CodeSynthesisConvertRule)
            ]

        if not self.allow_token_reduction:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, TokenReducedConvertRule)
            ]

        if not self.allow_rag_reduction:
            self.implementation_rules = [
                rule for rule in self.implementation_rules if not issubclass(rule, RAGConvertRule)
            ]

        if not self.allow_mixtures:
            self.implementation_rules = [
                rule for rule in self.implementation_rules
                if not issubclass(rule, MixtureOfAgentsConvertRule)
            ]

    def update_cost_model(self, cost_model: CostModel):
        self.cost_model = cost_model

    def get_physical_op_params(self):
        return {
            "verbose": self.verbose,
            "available_models": self.available_models,
            "champion_model": get_champion_model(self.available_models),
            "code_champion_model": get_code_champion_model(self.available_models),
            "conventional_fallback_model": get_conventional_fallback_model(self.available_models),
        }

    def construct_group_tree(self, dataset_nodes: list[Set]) -> tuple[list[int], dict[str, Field], dict[str, set[str]]]:
        # get node, output_schema, and input_schema(if applicable)
        node = dataset_nodes[-1]
        output_schema = node.schema
        input_schema = dataset_nodes[-2].schema if len(dataset_nodes) > 1 else None

        ### convert node --> Group ###
        uid = node.universal_identifier()

        # create the op for the given node
        op: LogicalOperator | None = None
        if not self.no_cache and DataDirectory().has_cached_answer(uid):
            op = CacheScan(dataset_id=uid, input_schema=None, output_schema=output_schema)
        elif isinstance(node, DataSource):
            op = BaseScan(dataset_id=uid, output_schema=output_schema)
        elif node._filter is not None:
            op = FilteredScan(
                input_schema=input_schema,
                output_schema=output_schema,
                filter=node._filter,
                depends_on=node._depends_on,
                target_cache_id=uid,
            )
        elif node._group_by is not None:
            op = GroupByAggregate(
                input_schema=input_schema,
                output_schema=output_schema,
                group_by_sig=node._group_by,
                target_cache_id=uid,
            )
        elif node._agg_func is not None:
            op = Aggregate(
                input_schema=input_schema,
                output_schema=output_schema,
                agg_func=node._agg_func,
                target_cache_id=uid,
            )
        elif node._limit is not None:
            op = LimitScan(
                input_schema=input_schema,
                output_schema=output_schema,
                limit=node._limit,
                target_cache_id=uid,
            )
        elif node._project_cols is not None:
            op = Project(
                input_schema=input_schema,
                output_schema=output_schema,
                project_cols=node._project_cols,
                target_cache_id=uid,
            )
        elif node._index is not None:
            op = RetrieveScan(
                input_schema=input_schema,
                output_schema=output_schema,
                index=node._index,
                search_attr=node._search_attr,
                output_attr=node._output_attr,
                k=node._k,
                target_cache_id=uid
            )
        elif output_schema != input_schema:
            op = ConvertScan(
                input_schema=input_schema,
                output_schema=output_schema,
                cardinality=node._cardinality,
                udf=node._udf,
                depends_on=node._depends_on,
                target_cache_id=uid,
            )
        else:
            raise NotImplementedError(
                f"""No logical operator exists for the specified dataset construction.
                {input_schema}->{output_schema} {"with filter:'" + node._filter + "'" if node._filter is not None else ""}"""
            )

        # compute the input group ids and fields for this node
        input_group_ids, input_group_fields, input_group_properties = (
            self.construct_group_tree(dataset_nodes[:-1]) if len(dataset_nodes) > 1 else ([], {}, {})
        )

        # compute the fields added by this operation and all fields
        input_group_short_field_names = list(map(lambda full_field: full_field.split(".")[-1], input_group_fields.keys()))
        new_fields = {
            field_name: field
            for field_name, field in op.output_schema.field_map(unique=True, id=uid).items()
            if (field_name.split(".")[-1] not in input_group_short_field_names) or (node._udf is not None)
        }
        all_fields = {**input_group_fields, **new_fields}

        # compute the set of (short) field names this operation depends on
        depends_on_field_names = (
            {}
            if isinstance(node, DataSource)
            else {field_name.split(".")[-1] for field_name in node._depends_on}
        )

        # compute all properties including this operations'
        all_properties = deepcopy(input_group_properties)
        if isinstance(op, FilteredScan):
            # NOTE: we could use op.get_op_id() here, but storing filter strings makes
            #       debugging a bit easier as you can read which filters are in the Group
            op_filter_str = op.filter.get_filter_str()
            if "filters" in all_properties:
                all_properties["filters"].add(op_filter_str)
            else:
                all_properties["filters"] = set([op_filter_str])

        elif isinstance(op, LimitScan):
            op_limit_str = op.get_logical_op_id()
            if "limits" in all_properties:
                all_properties["limits"].add(op_limit_str)
            else:
                all_properties["limits"] = set([op_limit_str])
    
        elif isinstance(op, Project):
            op_project_str = op.get_logical_op_id()
            if "projects" in all_properties:
                all_properties["projects"].add(op_project_str)
            else:
                all_properties["projects"] = set([op_project_str])

        # construct the logical expression and group
        logical_expression = LogicalExpression(
            operator=op,
            input_group_ids=input_group_ids,
            input_fields=input_group_fields,
            depends_on_field_names=depends_on_field_names,
            generated_fields=new_fields,
            group_id=None,
        )
        group = Group(
            logical_expressions=[logical_expression],
            fields=all_fields,
            properties=all_properties,
        )
        logical_expression.set_group_id(group.group_id)

        # add the expression and group to the optimizer's expressions and groups and return
        self.expressions[logical_expression.get_expr_id()] = logical_expression
        self.groups[group.group_id] = group

        return [group.group_id], all_fields, all_properties

    def convert_query_plan_to_group_tree(self, query_plan: Dataset) -> str:
        # Obtain ordered list of datasets
        dataset_nodes = []
        node = query_plan.copy()

        while isinstance(node, Dataset):
            dataset_nodes.append(node)
            node = node._source
        dataset_nodes.append(node)
        dataset_nodes = list(reversed(dataset_nodes))

        # remove unnecessary convert if output schema from data source scan matches
        # input schema for the next operator
        if len(dataset_nodes) > 1 and dataset_nodes[0].schema.get_desc() == dataset_nodes[1].schema.get_desc():
            dataset_nodes = [dataset_nodes[0]] + dataset_nodes[2:]
            if len(dataset_nodes) > 1:
                dataset_nodes[1]._source = dataset_nodes[0]

        # compute depends_on field for every node
        short_to_full_field_name = {}
        for node_idx, node in enumerate(dataset_nodes):
            # update mapping from short to full field names
            short_field_names = node.schema.field_names()
            full_field_names = node.schema.field_names(unique=True, id=node.universal_identifier())
            for short_field_name, full_field_name in zip(short_field_names, full_field_names):
                # set mapping automatically if this is a new field
                if short_field_name not in short_to_full_field_name or (
                    node_idx > 0 and dataset_nodes[node_idx - 1].schema != node.schema and node._udf is not None
                ):
                    short_to_full_field_name[short_field_name] = full_field_name

            # if the node is a data source, then skip
            if isinstance(node, DataSource):
                continue

            # If the node already has depends_on specified, then resolve each field name to a full (unique) field name
            if len(node._depends_on) > 0:
                node._depends_on = list(map(lambda field: short_to_full_field_name[field], node._depends_on))
                continue

            # otherwise, make the node depend on all upstream nodes
            node._depends_on = set()
            for upstream_node in dataset_nodes[:node_idx]:
                node._depends_on.update(upstream_node.schema.field_names(unique=True, id=upstream_node.universal_identifier()))
            node._depends_on = list(node._depends_on)

        # construct tree of groups
        final_group_id, _, _ = self.construct_group_tree(dataset_nodes)

        # check that final_group_id is a singleton
        assert len(final_group_id) == 1
        final_group_id = final_group_id[0]

        return final_group_id

    def heuristic_optimization(self, group_id: int) -> None:
        """
        Apply universally desirable transformations (e.g. filter/projection push-down).
        """
        pass

    def search_optimization_space(self, group_id: int) -> None:
        # begin the search for an optimal plan with a task to optimize the final group
        initial_task = OptimizeGroup(group_id)
        self.tasks_stack.append(initial_task)

        # TODO: conditionally stop when X number of tasks have been executed to limit exhaustive search
        while len(self.tasks_stack) > 0:
            task = self.tasks_stack.pop(-1)
            new_tasks = []
            if isinstance(task, (OptimizeGroup, ExpandGroup)):
                new_tasks = task.perform(self.groups)
            elif isinstance(task, OptimizeLogicalExpression):
                new_tasks = task.perform(self.transformation_rules, self.implementation_rules)
            elif isinstance(task, ApplyRule):
                context = {"costed_phys_op_ids": self.costed_phys_op_ids}
                new_tasks = task.perform(self.groups, self.expressions, context=context, **self.get_physical_op_params())
            elif isinstance(task, OptimizePhysicalExpression):
                context = {"optimization_strategy": self.optimization_strategy}
                new_tasks = task.perform(self.cost_model, self.groups, self.policy, context=context)

            self.tasks_stack.extend(new_tasks)

    def get_sentinel_plan(self, group_id: int) -> SentinelPlan:
        """
        Create and return a SentinelPlan object.
        """
        # get all the physical expressions for this group
        phys_exprs = self.groups[group_id].physical_expressions
        phys_op_set = [expr.operator for expr in phys_exprs]

        # if this expression has no inputs (i.e. it is a BaseScan or CacheScan),
        # create and return the physical plan
        best_phys_expr = self.groups[group_id].best_physical_expression
        if len(best_phys_expr.input_group_ids) == 0:
            return SentinelPlan(operator_sets=[phys_op_set])

        # TODO: need to handle joins
        # get the best physical plan(s) for this group's inputs
        best_phys_subplan = SentinelPlan(operator_sets=[])
        for input_group_id in best_phys_expr.input_group_ids:
            input_best_phys_plan = self.get_sentinel_plan(input_group_id)
            best_phys_subplan = SentinelPlan.from_ops_and_sub_plan(best_phys_subplan.operator_sets, input_best_phys_plan)

        # add this operator set to best physical plan and return
        return SentinelPlan.from_ops_and_sub_plan([phys_op_set], best_phys_subplan)

    def get_greedy_physical_plan(self, group_id: int) -> PhysicalPlan:
        """
        Return the best plan with respect to the user provided policy.
        """
        # get the best physical expression for this group
        best_phys_expr = self.groups[group_id].best_physical_expression

        # if this expression has no inputs (i.e. it is a BaseScan or CacheScan),
        # create and return the physical plan
        if len(best_phys_expr.input_group_ids) == 0:
            return PhysicalPlan(operators=[best_phys_expr.operator], plan_cost=best_phys_expr.plan_cost)

        # get the best physical plan(s) for this group's inputs
        input_group_id = best_phys_expr.input_group_ids[0] # TODO: need to handle joins
        input_best_phys_plan = self.get_greedy_physical_plan(input_group_id)

        # add this operator to best physical plan and return
        return PhysicalPlan.from_ops_and_sub_plan([best_phys_expr.operator], input_best_phys_plan, best_phys_expr.plan_cost)


    def get_candidate_pareto_physical_plans(self, group_id: int, policy: Policy) -> list[PhysicalPlan]:
        """
        Return a list of plans which will contain all of the pareto optimal plans (and some additional
        plans which may not be pareto optimal).

        TODO: can we cache group_id --> final_pareto_optimal_plans to avoid re-computing upstream
        groups' pareto-optimal plans for each expression?
        """
        # get the pareto optimal physical expressions for this group
        pareto_optimal_phys_exprs = self.groups[group_id].pareto_optimal_physical_expressions

        # construct list of pareto optimal plans
        pareto_optimal_plans = []
        for phys_expr in pareto_optimal_phys_exprs:
            # if this expression has no inputs (i.e. it is a BaseScan or CacheScan),
            # create and return the physical plan
            if len(phys_expr.input_group_ids) == 0:
                for plan_cost, _ in phys_expr.pareto_optimal_plan_costs:
                    plan = PhysicalPlan(operators=[phys_expr.operator], plan_cost=plan_cost)
                    pareto_optimal_plans.append(plan)

            # otherwise, get the pareto optimal physical plan(s) for this group's inputs
            else:
                # get the pareto optimal physical plan(s) for this group's inputs
                input_group_id = phys_expr.input_group_ids[0] # TODO: need to handle joins
                pareto_optimal_phys_subplans = self.get_candidate_pareto_physical_plans(input_group_id, policy)

                # iterate over the input subplans and find the one(s) which combine with this physical expression
                # to make a pareto-optimal plan
                for plan_cost, input_plan_cost in phys_expr.pareto_optimal_plan_costs:
                    for subplan in pareto_optimal_phys_subplans:
                        if (
                            subplan.plan_cost.cost == input_plan_cost.cost
                            and subplan.plan_cost.time == input_plan_cost.time
                            and subplan.plan_cost.quality == input_plan_cost.quality
                        ):
                            # TODO: The plan_cost gets summed with subplan.plan_cost;
                            #       am I defining expression.best_plan_cost to be the cost of that operator,
                            #       and expression.pareto_optimal_plan_costs to be the cost(s) of the subplan including that operator?
                            #       i.e. are my definitions inconsistent?
                            plan = PhysicalPlan.from_ops_and_sub_plan([phys_expr.operator], subplan, plan_cost)
                            pareto_optimal_plans.append(plan)

        return pareto_optimal_plans

    def get_confidence_interval_optimal_plans(self, group_id: int) -> list[PhysicalPlan]:
        """
        Return all physical plans whose upper bound on the primary policy metric is greater than the
        best plan's lower bound on the primary policy metric (subject to satisfying the policy constraint).

        The OptimizePhysicalExpression task guarantees that each group's `ci_best_physical_expressions`
        maintains a list of expressions with overlapping CI's on the primary policy metric (while also
        satisfying the policy constraint).

        This function computes the cross-product of all such expressions across all groups.
        """
        # get all the physical expressions which could be the best for this group
        best_phys_exprs = self.groups[group_id].ci_best_physical_expressions

        best_plans = []
        for phys_expr in best_phys_exprs:
            # if this expression has no inputs (i.e. it is a BaseScan or CacheScan),
            # create the physical plan and append it to the best_plans for this group
            if len(phys_expr.input_group_ids) == 0:
                plan = PhysicalPlan(operators=[phys_expr.operator], plan_cost=phys_expr.plan_cost)
                best_plans.append(plan)

            # otherwise, get the best physical plan(s) for this group's inputs
            else:
                # TODO: need to handle joins
                best_phys_subplans = [PhysicalPlan(operators=[])]
                for input_group_id in phys_expr.input_group_ids:
                    input_best_phys_plans = self.get_confidence_interval_optimal_plans(input_group_id)
                    best_phys_subplans = [
                        PhysicalPlan.from_ops_and_sub_plan(subplan.operators, input_subplan, subplan.plan_cost)
                        for subplan in best_phys_subplans
                        for input_subplan in input_best_phys_plans
                    ]

                # add this operator to best physical plan and return
                for subplan in best_phys_subplans:
                    plan = PhysicalPlan.from_ops_and_sub_plan([phys_expr.operator], subplan, phys_expr.plan_cost)
                    best_plans.append(plan)

        return best_plans


    def optimize(self, query_plan: Dataset, policy: Policy | None = None) -> list[PhysicalPlan]:
        """
        The optimize function takes in an initial query plan and searches the space of
        logical and physical plans in order to cost and produce a (near) optimal physical plan.
        """
        # compute the initial group tree for the user plan
        final_group_id = self.convert_query_plan_to_group_tree(query_plan)

        # TODO
        # # do heuristic based pre-optimization
        # self.heuristic_optimization(final_group_id)

        # search the optimization space by applying logical and physical transformations to the initial group tree
        self.search_optimization_space(final_group_id)

        # construct the optimal physical plan(s) by traversing the memo table
        plans = []
        if self.optimization_strategy == OptimizationStrategy.SENTINEL:
            plans = [self.get_sentinel_plan(final_group_id)]

        elif self.optimization_strategy == OptimizationStrategy.GREEDY:
            plans = [self.get_greedy_physical_plan(final_group_id)]

        elif self.optimization_strategy == OptimizationStrategy.PARETO:
            # compute all of the pareto optimal physical plans
            plans = self.get_candidate_pareto_physical_plans(final_group_id, policy)

            # adjust plans' plan_cost.quality to reflect only the quality of the final operator
            if self.use_final_op_quality:
                for plan in plans:
                    plan.plan_cost.quality = plan.plan_cost.op_estimates.quality

            # filter pareto optimal plans for ones which satisfy policy constraint (if at least one of them does)
            # import pdb; pdb.set_trace()
            if any([policy.constraint(plan.plan_cost) for plan in plans]):
                plans = [plan for plan in plans if policy.constraint(plan.plan_cost)]

            # select the plan which is best for the given policy
            optimal_plan, plans = plans[0], plans[1:]
            for plan in plans:
                optimal_plan = optimal_plan if policy.choose(optimal_plan.plan_cost, plan.plan_cost) else plan

            plans = [optimal_plan]

        elif self.optimization_strategy == OptimizationStrategy.CONFIDENCE_INTERVAL:
            # TODO: fix this to properly handle multiple potential plans
            raise Exception("NotImplementedError")
            plans = self.get_confidence_interval_optimal_plans(final_group_id)

        elif self.optimization_strategy == OptimizationStrategy.NONE:
            plans = [self.get_greedy_physical_plan(final_group_id)]

        return plans



import time

from palimpzest.core.data.dataclasses import OperatorStats, PlanStats
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.lib.schemas import SourceRecord
from palimpzest.policy import Policy
from palimpzest.query.execution.execution_engine import ExecutionEngine
from palimpzest.query.operators.aggregate import AggregateOp
from palimpzest.query.operators.datasource import DataSourcePhysicalOp, MarshalAndScanDataOp
from palimpzest.query.operators.filter import FilterOp
from palimpzest.query.operators.limit import LimitScanOp
from palimpzest.query.optimizer.cost_model import CostModel
from palimpzest.query.optimizer.optimizer import Optimizer
from palimpzest.query.optimizer.plan import PhysicalPlan
from palimpzest.sets import Dataset


class StreamingSequentialExecution(ExecutionEngine):
    """This class can be used for a streaming, record-based execution.
    Results are returned as an iterable that can be consumed by the caller."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._plan: PhysicalPlan | None = None
        self._plan_stats: PlanStats | None = None
        self.last_record = False
        self.current_scan_idx: int = 0
        self.plan_generated: bool = False
        self.records_count: int = 0

    @property
    def plan(self) -> PhysicalPlan:
        if self._plan is None:
            raise Exception("Plan has not been generated yet.")
        return self._plan

    @plan.setter
    def plan(self, plan: PhysicalPlan):
        self._plan = plan

    @property
    def plan_stats(self) -> PlanStats:
        if self._plan_stats is None:
            raise Exception("Plan stats have not been generated yet.")
        return self._plan_stats

    @plan_stats.setter
    def plan_stats(self, plan_stats: PlanStats):
        self._plan_stats = plan_stats

    def generate_plan(self, dataset: Dataset, policy: Policy):
        self.clear_cached_examples()
        start_time = time.time()

        cost_model = CostModel()
        optimizer = Optimizer(
            policy=policy,
            cost_model=cost_model,
            no_cache=self.nocache,
            verbose=self.verbose,
            available_models=self.available_models,
            allow_bonded_query=self.allow_bonded_query,
            allow_conventional_query=self.allow_conventional_query,
            allow_code_synth=self.allow_code_synth,
            allow_token_reduction=self.allow_token_reduction,
            allow_rag_reduction=self.allow_rag_reduction,
            allow_mixtures=self.allow_mixtures,
            optimization_strategy=self.optimization_strategy,
        )

        # Effectively always use the optimal strategy
        plans = optimizer.optimize(dataset, policy)
        self.plan = plans[0]
        self.plan_stats = PlanStats(plan_id=self.plan.plan_id)
        for op in self.plan.operators:
            if isinstance(op, AggregateOp):
                raise Exception("You cannot have a Streaming Execution if there is an Aggregation Operator")
            op_id = op.get_op_id()
            op_name = op.op_name()
            op_details = {k: str(v) for k, v in op.get_id_params().items()}
            self.plan_stats.operator_stats[op_id] = OperatorStats(op_id=op_id, op_name=op_name, op_details=op_details) 
        print("Time for planning: ", time.time() - start_time)
        self.plan_generated = True
        return self.plan

    def execute(
        self,
        dataset: Dataset,
        policy: Policy,
    ):
        start_time = time.time()
        # Always delete cache
        if not self.plan_generated:
            self.generate_plan(dataset, policy)

        input_records = self.get_input_records()
        for idx, record in enumerate(input_records):
            # print("Iteration number: ", idx+1, "out of", len(input_records))
            output_records = self.execute_opstream(self.plan, record)
            if idx == len(input_records) - 1:
                total_plan_time = time.time() - start_time
                self.plan_stats.finalize(total_plan_time)

            yield output_records, self.plan, self.plan_stats

    def get_input_records(self):
        scan_operator = self.plan.operators[0]
        datasource = (
            self.datadir.get_registered_dataset(scan_operator.dataset_id)
            if isinstance(scan_operator, MarshalAndScanDataOp)
            else self.datadir.get_cached_result(scan_operator.dataset_id)
        )
        if not datasource:
            raise Exception("Data source not found")
        datasource_len = len(datasource)

        input_records = []
        record_op_stats = []
        for idx in range(datasource_len):
            # NOTE: this DataRecord will be discarded and replaced by the scan_operator;
            #       it is simply a vessel to inform the scan_operator which record to fetch
            candidate = DataRecord(schema=SourceRecord, source_id=idx)
            candidate.idx = idx
            candidate.get_item_fn = datasource.get_item
            records, record_op_stats_lst = scan_operator(candidate)
            input_records += records
            record_op_stats += record_op_stats_lst

        op_id = scan_operator.get_op_id()
        self.plan_stats.operator_stats[op_id].add_record_op_stats(
            record_op_stats,
            source_op_id=None,
            plan_id=self.plan.plan_id,
        )

        return input_records

    def execute_opstream(self, plan, record):
        # initialize list of output records and intermediate variables
        input_records = [record]
        record_op_stats_lst = []

        for op_idx, operator in enumerate(plan.operators):
            # TODO: this being defined in the for loop potentially makes the return
            # unbounded if plan.operators is empty. This should be defined outside the loop
            # and the loop refactored to account for not redeclaring this for each operator
            output_records = []
            op_id = operator.get_op_id()
            prev_op_id = plan.operators[op_idx - 1].get_op_id() if op_idx > 1 else None

            if isinstance(operator, DataSourcePhysicalOp):
                continue
            # only invoke aggregate operator(s) once there are no more source records and all
            # upstream operators' processing queues are empty
            # elif isinstance(operator, AggregateOp):
            # output_records, record_op_stats_lst = operator(candidates=input_records)
            elif isinstance(operator, LimitScanOp):
                if self.records_count >= operator.limit:
                    break
            else:
                for r in input_records:
                    record_out, stats = operator(r)
                    output_records += record_out
                    record_op_stats_lst += stats

                if isinstance(operator, FilterOp):
                    # delete all records that did not pass the filter
                    output_records = [r for r in output_records if r.passed_operator]
                    if not output_records:
                        break

            self.plan_stats.operator_stats[op_id].add_record_op_stats(
                record_op_stats_lst,
                source_op_id=prev_op_id,
                plan_id=plan.plan_id,
            )
            input_records = output_records
            self.records_count += len(output_records)

        return output_records



import os
from typing import List, Optional

from palimpzest.constants import Model


def get_vision_models() -> List[Model]:
    """
    Return the set of vision models which the system has access to based on the set of environment variables.
    """
    models = []
    if os.getenv("OPENAI_API_KEY") is not None:
        models.extend([Model.GPT_4o_V, Model.GPT_4o_MINI_V])

    if os.getenv("TOGETHER_API_KEY") is not None:
        models.extend([Model.LLAMA3_V])

    return models


def get_models(include_vision: Optional[bool] = False) -> List[Model]:
    """
    Return the set of models which the system has access to based on the set environment variables.
    """
    models = []
    if os.getenv("OPENAI_API_KEY") is not None:
        models.extend([Model.GPT_4o, Model.GPT_4o_MINI])

    if os.getenv("TOGETHER_API_KEY") is not None:
        models.extend([Model.LLAMA3, Model.MIXTRAL])

    if include_vision:
        vision_models = get_vision_models()
        models.extend(vision_models)

    return models


def get_champion_model(available_models, vision=False):
    champion_model = None

    # non-vision
    if not vision and Model.GPT_4o in available_models:
        champion_model = Model.GPT_4o
    elif not vision and Model.GPT_4o_MINI in available_models:
        champion_model = Model.GPT_4o_MINI
    elif not vision and Model.LLAMA3 in available_models:
        champion_model = Model.LLAMA3
    elif not vision and Model.MIXTRAL in available_models:
        champion_model = Model.MIXTRAL

    # vision
    elif vision and Model.GPT_4o_V in available_models:
        champion_model = Model.GPT_4o_V
    elif vision and Model.GPT_4o_MINI_V in available_models:
        champion_model = Model.GPT_4o_MINI_V
    elif vision and Model.LLAMA3_V in available_models:
        champion_model = Model.LLAMA3_V

    else:
        raise Exception(
            "No models available to create physical plans! You must set at least one of the following environment"
            " variables: [OPENAI_API_KEY, TOGETHER_API_KEY, GOOGLE_API_KEY]\n"
            f"available_models: {available_models}"
        )

    return champion_model


def get_conventional_fallback_model(available_models, vision=False):
    return get_champion_model(available_models, vision)


def get_code_champion_model(available_models):
    # NOTE: for now, assume same champion as get_champion_model()
    return get_champion_model(available_models, vision=False)


def get_champion_model_name(available_models, vision=False):
    return get_champion_model(available_models, vision).value



import os
import sys
import tempfile

import yaml

from palimpzest.constants import PZ_DIR


class Config:
    def __init__(self, name: str = "default", llmservice: str = "openai", parallel: bool = False):
        self.config_file_path = os.path.join(PZ_DIR, f"config_{name}.yaml")
        if not os.path.exists(PZ_DIR):
            raise Exception(
                f"Target config directory does not exist at {PZ_DIR} :: Something is wrong with the installation."
            )

        if not os.path.exists(self.config_file_path):
            # Get the system's temporary directory
            temp_dir = tempfile.gettempdir()
            pz_file_cache_dir = os.path.join(temp_dir, "pz")
            self.config = {
                "name": name,
                "llmservice": llmservice,
                "parallel": parallel,
                "filecachedir": pz_file_cache_dir,
            }
            self._save_config()

        self.config = self._load_config()
        if "filecachedir" not in self.config:
            # Get the system's temporary directory
            temp_dir = tempfile.gettempdir()
            pz_file_cache_dir = os.path.join(temp_dir, "pz")
            self.config["filecachedir"] = pz_file_cache_dir
            self._save_config()

        self.name = self.config["name"]

    def get(self, key, default=None):
        return self.config.get(key, default)

    def set(self, key, value):
        self.config[key] = value
        self._save_config()

    def set_current_config(self):
        current_config_dict = {"current_config_name": self.name}
        current_config_path = os.path.join(PZ_DIR, "current_config.yaml")
        with open(current_config_path, "w") as f:
            yaml.dump(current_config_dict, f)

    def remove_config(self):
        # check to ensure you don't delete default config
        if self.name == "default":
            raise Exception("Cannot remove default config.")

        # reset current config if this config was the current config
        current_config_path = os.path.join(PZ_DIR, "current_config.yaml")
        current_config_dict = {}
        with open(current_config_path) as f:
            current_config_dict = yaml.safe_load(f)

        if current_config_dict["current_config_name"] == self.name:
            current_config_dict["current_config_name"] = "default"

            with open(current_config_path, "w") as f:
                yaml.dump(current_config_dict, f)

        # delete config file
        os.remove(self.config_file_path)

    def _load_config(self):
        """Load YAML configuration from the specified path."""
        try:
            with open(self.config_file_path) as file:
                return yaml.safe_load(file)
        except Exception as e:
            print(f"Error loading configuration file: {e}")
            sys.exit(1)

    def _save_config(self):
        """Save the configuration to the specified path."""
        try:
            with open(self.config_file_path, "w") as file:
                yaml.dump(self.config, file)
        except Exception as e:
            print(f"Error saving configuration file: {e}")
            sys.exit(1)



import multiprocessing
from concurrent.futures import ThreadPoolExecutor

from palimpzest.constants import Model, OptimizationStrategy
from palimpzest.core.data.dataclasses import PlanStats, RecordOpStats
from palimpzest.core.data.datasources import DataSource, ValidationDataSource
from palimpzest.core.elements.records import DataRecord
from palimpzest.datamanager.datamanager import DataDirectory
from palimpzest.policy import Policy
from palimpzest.query.optimizer.cost_model import CostModel
from palimpzest.query.optimizer.optimizer import Optimizer
from palimpzest.query.optimizer.plan import PhysicalPlan
from palimpzest.sets import Dataset, Set
from palimpzest.utils.hash_helpers import hash_for_id
from palimpzest.utils.model_helpers import get_models


class ExecutionEngine:
    def __init__(
        self,
        datasource: DataSource,
        num_samples: int = float("inf"),
        scan_start_idx: int = 0,
        nocache: bool = True,  # NOTE: until we properly implement caching, let's set the default to True
        include_baselines: bool = False,
        min_plans: int | None = None,
        verbose: bool = False,
        available_models: list[Model] | None = None,
        allow_bonded_query: bool = True,
        allow_conventional_query: bool = False,
        allow_model_selection: bool = True,
        allow_code_synth: bool = True,
        allow_token_reduction: bool = False,
        allow_rag_reduction: bool = True,
        allow_mixtures: bool = True,
        optimization_strategy: OptimizationStrategy = OptimizationStrategy.PARETO,
        max_workers: int | None = None,
        num_workers_per_plan: int = 1,
        *args,
        **kwargs,
    ) -> None:
        self.num_samples = num_samples
        self.scan_start_idx = scan_start_idx
        self.nocache = nocache
        if not self.nocache:
            raise NotImplementedError("Caching is not yet implemented! Please set nocache=True.")
        self.include_baselines = include_baselines
        self.min_plans = min_plans
        self.verbose = verbose
        self.available_models = available_models
        if self.available_models is None or len(self.available_models) == 0:
            self.available_models = get_models(include_vision=True)
        if self.verbose:
            print("Available models: ", self.available_models)
        self.allow_bonded_query = allow_bonded_query
        self.allow_conventional_query = allow_conventional_query
        self.allow_model_selection = allow_model_selection
        self.allow_code_synth = allow_code_synth
        self.allow_token_reduction = allow_token_reduction
        self.allow_rag_reduction = allow_rag_reduction
        self.allow_mixtures = allow_mixtures
        self.optimization_strategy = optimization_strategy
        self.max_workers = max_workers
        self.num_workers_per_plan = num_workers_per_plan

        self.datadir = DataDirectory()

        # datasource; should be set by execute() with call to get_datasource()
        self.datasource = datasource
        self.using_validation_data = isinstance(self.datasource, ValidationDataSource)


    def execution_id(self) -> str:
        """
        Hash of the class parameters.
        """
        id_str = ""
        for attr, value in self.__dict__.items():
            if not attr.startswith("_"):
                id_str += f"{attr}={value},"

        return hash_for_id(id_str)

    def clear_cached_examples(self):
        """
        Clear cached codegen samples.
        """
        cache = self.datadir.get_cache_service()
        cache.rm_cache()

    def get_parallel_max_workers(self):
        # for now, return the number of system CPUs;
        # in the future, we may want to consider the models the user has access to
        # and whether or not they will encounter rate-limits. If they will, we should
        # set the max workers in a manner that is designed to avoid hitting them.
        # Doing this "right" may require considering their logical, physical plan,
        # and tier status with LLM providers. It may also be worth dynamically
        # changing the max_workers in response to 429 errors.
        return max(int(0.8 * multiprocessing.cpu_count()), 1)

    def get_max_quality_plan_id(self, plans: list[PhysicalPlan]) -> str:
        """
        Return the plan_id for the plan with the highest quality in the list of plans.
        """
        max_quality_plan_id, max_quality = None, -1
        for plan in plans:
            if plan.quality > max_quality or max_quality_plan_id is None:
                max_quality_plan_id = plan.plan_id
                max_quality = plan.quality

        return max_quality_plan_id

    def aggregate_plan_stats(self, plan_stats: list[PlanStats]) -> dict[str, PlanStats]:
        """
        Aggregate a list of plan stats into a dictionary mapping plan_id --> cumulative plan stats.

        NOTE: we make the assumption that the same plan cannot be run more than once in parallel,
        i.e. each plan stats object for an individual plan comes from two different (sequential)
        periods in time. Thus, PlanStats' total_plan_time(s) can be summed.
        """
        agg_plan_stats = {}
        for ps in plan_stats:
            if ps.plan_id in agg_plan_stats:
                agg_plan_stats[ps.plan_id] += ps
            else:
                agg_plan_stats[ps.plan_id] = ps

        return agg_plan_stats

    def execute_plans(
        self, plans: list[PhysicalPlan], max_quality_plan_id: str, num_samples: int | float = float("inf")
    ):
        """
        Execute a given list of plans for num_samples records each. Plans are executed in parallel.
        If any workers are unused, then additional workers are distributed evenly among plans.
        """
        # compute number of plans
        num_plans = len(plans)

        # set plan_parallel_workers and workers_per_plan;
        # - plan_parallel_workers controls how many plans are executed in parallel
        # - workers_per_plan controls how many threads are assigned to executing each plan
        plan_parallel_workers, workers_per_plan = None, None
        if self.max_workers <= num_plans:
            plan_parallel_workers = self.max_workers
            workers_per_plan = [1 for _ in range(num_plans)]
        else:
            plan_parallel_workers = num_plans
            workers_per_plan = [(self.max_workers // num_plans) for _ in range(num_plans)]
            idx = 0
            while sum(workers_per_plan) < self.max_workers:
                workers_per_plan[idx] += 1
                idx += 1

        with ThreadPoolExecutor(max_workers=plan_parallel_workers) as executor:
            results = list(executor.map(lambda x: self.execute_plan(**x),
                    [{"plan": plan,
                      "num_samples": num_samples,
                      "plan_workers": plan_workers}
                      for plan, plan_workers in zip(plans, workers_per_plan)],
                )
            )
        # results = list(map(lambda x: self.execute_plan(**x),
        #         [{"plan": plan,
        #             "num_samples": num_samples,
        #             "plan_workers": 1}
        #             for plan in plans],
        #     )
        # )
        # split results into per-plan records and plan stats
        all_records, all_plan_stats = zip(*results)

        # process results to get sample execution data and sentinel plan stats
        all_sample_execution_data, return_records = [], []
        for records, plan_stats, plan in zip(all_records, all_plan_stats, plans):
            # aggregate sentinel est. data
            for operator_stats in plan_stats.operator_stats.values():
                all_sample_execution_data.extend(operator_stats.record_op_stats_lst)

            # if this is the max quality plan for this set of plans, return its results for these records
            if plan.plan_id == max_quality_plan_id:
                return_records = records

        return all_sample_execution_data, return_records, all_plan_stats

    def execute_strategy(
        self,
        dataset: Set,
        policy: Policy,
        optimizer: Optimizer,
        execution_data: list[RecordOpStats] | None = None,
    ) -> tuple[list[DataRecord], list[PlanStats]]:
        if execution_data is None:
            execution_data = []

        # get the optimal plan according to the optimizer
        plans = optimizer.optimize(dataset, policy)
        final_plan = plans[0]

        # execute the plan
        # TODO: for some reason this is not picking up change to self.max_workers from PipelinedParallelPlanExecutor.__init__()
        records, plan_stats = self.execute_plan(
            plan=final_plan,
            plan_workers=self.max_workers,
        )

        # return the output records and plan stats
        return records, [plan_stats]

    def execute_confidence_interval_strategy(
        self,
        dataset: Set,
        policy: Policy,
        optimizer: Optimizer,
        execution_data: list[RecordOpStats] | None = None,
    ) -> tuple[list[DataRecord], list[PlanStats]]:
        # initialize output records and plan stats
        if execution_data is None:
            execution_data = []
        records, plan_stats = [], []

        # get the initial set of optimal plans according to the optimizer
        plans = optimizer.optimize(dataset, policy)
        while len(plans) > 1 and self.scan_start_idx < len(self.datasource):
            # identify the plan with the highest quality in the set
            max_quality_plan_id = self.get_max_quality_plan_id(plans)

            # execute the set of plans for a fixed number of samples
            new_execution_data, new_records, new_plan_stats = self.execute_plans(
                list(plans), max_quality_plan_id, self.num_samples
            )
            records.extend(new_records)
            plan_stats.extend(new_plan_stats)

            if self.scan_start_idx + self.num_samples < len(self.datasource):
                # update cost model and optimizer
                execution_data.extend(new_execution_data)
                cost_model = CostModel(sample_execution_data=execution_data)
                optimizer.update_cost_model(cost_model)

                # get new set of plans
                plans = optimizer.optimize(dataset, policy)

                # update scan start idx
                self.scan_start_idx += self.num_samples

        if self.scan_start_idx < len(self.datasource):
            # execute final plan until end
            final_plan = plans[0]
            new_records, new_plan_stats = self.execute_plan(
                plan=final_plan,
                plan_workers=self.max_workers,
            )
            records.extend(new_records)
            plan_stats.append(new_plan_stats)

        # return the final set of records and plan stats
        return records, plan_stats


    def execute_plan(self, plan: PhysicalPlan, num_samples: int | float = float("inf"), plan_workers: int = 1):
        """Execute the given plan and return the output records and plan stats."""
        raise NotImplementedError("Abstract method to be overwritten by sub-classes")


    def execute(self, dataset: Dataset, policy: Policy):
        """
        Execute the workload specified by the given dataset according to the policy provided by the user.
        """
        raise NotImplementedError("Abstract method to be overwritten by sub-classes")



from __future__ import annotations

import math
from typing import Any

from palimpzest.constants import (
    MODEL_CARDS,
    NAIVE_EST_NUM_INPUT_TOKENS,
    NAIVE_EST_NUM_OUTPUT_TOKENS,
)
from palimpzest.core.data.dataclasses import OperatorCostEstimates
from palimpzest.query.operators.convert import LLMConvert, LLMConvertBonded, LLMConvertConventional
from palimpzest.utils.token_reduction_helpers import best_substring_match, find_best_range


# NOTE: this convert operation will not work with the new generation abstraction, and it needs to be worked on.
#       There are two minor issues with the operator as it exists:
#
#       1) The token reduction operation operated over the entire JSON string of the input DataRecord
#          - while this works in practice, it makes it difficult to use this operator with a generation framework
#            where each field may be placed in a specific place in the format string for a prompt
#          - we need to either (A) rewrite the reduction to take place on a field-by-field basis (or at least
#            make it possible to recover each field after a global reduction) or (B) add custom logic within
#            the Generator class(es) to handle this operator [I much prefer (A) over (B)]
#
#       2) The heatmap update logic does not translate well to the distributed setting, where this operator may
#          be copied and executed many times in parallel
#          - each copy of the operator will have its own heatmap and require MAX_HEATMAP_UPDATES just to enter the
#            phase where token reduction takes place
#          - this means that if we have 20-way parallelism and a MAX_HEATMAP_UPDATES = 5, it can take 100 inputs
#            before token reduction ever takes place
#          - this also creates difficulties in properly performing cost-estimation for this operator; e.g. if we use
#            n <= MAX_HEATMAP_UPDATES samples to cost this operator, then we will never actually measure its performance
#            in the token reduction phase -- which could have a serious degradation in quality that our optimizer doesn't see
class TokenReducedConvert(LLMConvert):
    # NOTE: moving these closer to the TokenReducedConvert class for now (in part to make
    #       them easier to mock); we can make these parameterized as well
    MAX_HEATMAP_UPDATES: int = 5
    TOKEN_REDUCTION_SAMPLE: int = 0
    TOKEN_REDUCTION_GRANULARITY: float = 0.001

    def __init__(self, token_budget: float, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.token_budget = token_budget
        self.resolution = self.TOKEN_REDUCTION_GRANULARITY
        self.first_execution = True
        self.count = 0
        self.heatmap = [0] * int(1.0 / self.resolution)

    def __str__(self):
        op = super().__str__()
        op += f"    Token Budget: {str(self.token_budget)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {"token_budget": self.token_budget, **id_params}

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"token_budget": self.token_budget, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Update the cost per record and quality estimates produced by LLMConvert's naive estimates.
        We adjust the cost per record to account for the reduced number of input tokens following
        token reduction, and we make a crude estimate of the quality degradation that results from
        using fewer tokens.
        """
        # get naive cost estimates from LLMConvert
        naive_op_cost_estimates = super().naive_cost_estimates(source_op_cost_estimates)

        # re-compute cost per record assuming we use fewer input tokens
        est_num_input_tokens = NAIVE_EST_NUM_INPUT_TOKENS * self.token_budget
        est_num_output_tokens = NAIVE_EST_NUM_OUTPUT_TOKENS
        model_conversion_usd_per_record = (
            MODEL_CARDS[self.model.value]["usd_per_input_token"] * est_num_input_tokens
            + MODEL_CARDS[self.model.value]["usd_per_output_token"] * est_num_output_tokens
        )

        # set refined estimate of cost per record and, for now,
        # assume quality multiplier is proportional to sqrt(sqrt(token_budget))
        naive_op_cost_estimates.cost_per_record = model_conversion_usd_per_record
        naive_op_cost_estimates.cost_per_record_lower_bound = naive_op_cost_estimates.cost_per_record
        naive_op_cost_estimates.cost_per_record_upper_bound = naive_op_cost_estimates.cost_per_record
        naive_op_cost_estimates.quality = (naive_op_cost_estimates.quality) * math.sqrt(math.sqrt(self.token_budget))
        naive_op_cost_estimates.quality_lower_bound = naive_op_cost_estimates.quality
        naive_op_cost_estimates.quality_upper_bound = naive_op_cost_estimates.quality

        return naive_op_cost_estimates
    
    def is_image_conversion(self) -> bool:
        """TokenReducedConvert is currently disallowed on image conversions, so this must be False."""
        return False

    def reduce_context(self, full_context: str) -> str:
        range = find_best_range(
            self.heatmap,
            int(self.token_budget / self.TOKEN_REDUCTION_GRANULARITY),
            trim_zeros=False,
        )
        if not range:
            raise Exception("No range found in heatmap")
        si, ei = range
        print("si:", si, "ei:", ei)
        sr, er = (
            si * self.TOKEN_REDUCTION_GRANULARITY,
            ei * self.TOKEN_REDUCTION_GRANULARITY,
        )
        test_len = len(full_context)
        start = int(sr * test_len)
        end = int(er * test_len)
        if self.verbose:
            print(f"start ratio: {sr} -- end ratio: {er}")
            print("character start:", start, "end:", end)
        sample = full_context[start:end]
        return sample

    def _dspy_generate_fields(self, prompt: str, content: str | list[str]) -> tuple[list[dict[str, list]] | Any]:
        raise Exception("TokenReducedConvert is executing despite being deprecated until implementation changes can be made.")
        answer, query_stats = None, None
        if self.first_execution or self.count < self.MAX_HEATMAP_UPDATES:
            if self.verbose:
                print("Warming up heatmap")
            answer, query_stats = super()._dspy_generate_fields(prompt, content)
            self.first_execution = False

        else:
            if self.verbose:
                print("Using heatmap")

            # only refer to the heatmap if the count is greater than a enough sample size
            # TODO: only trim the context if the attention is clustered in a small region
            if self.count >= self.TOKEN_REDUCTION_SAMPLE:
                context = self.reduce_context(content)
                try:
                    answer, _, query_stats = self.generator.generate(context=context, prompt=prompt)
                except Exception as e:
                    print(f"DSPy generation error: {e}, falling back to unreduced generation")
                    answer, query_stats = super()._dspy_generate_fields(prompt, content)

        # TODO: answer and query stats may be unbound if we hit the else block
        # and count < TOKEN_REDUCTION_SAMPLE, which makes the below pretty clunky
        # this throw asserts our view of the world and we should refactor this
        if answer is None or query_stats is None:
            raise Exception("answer or query_stats is None")
        try:
            match = best_substring_match(answer, content)
            if not match:
                gsi, gei = 0, len(content)
            else:
                gsi, gei = match
        except Exception as e:
            print("Error in substring match:", e)
            gsi, gei = 0, len(content)
        context_len = len(content)
        gsr, ger = gsi / context_len, gei / context_len
        norm_si, norm_ei = int(gsr / self.resolution), int(ger / self.resolution)
        if self.verbose:
            print(f"best_start: {gsi} -- best_end: {gei}")

        self.count += 1
        self.heatmap[norm_si:norm_ei] = map(lambda x: x + 1, self.heatmap[norm_si:norm_ei])

        return answer, query_stats


class TokenReducedConvertConventional(TokenReducedConvert, LLMConvertConventional):
    pass


class TokenReducedConvertBonded(TokenReducedConvert, LLMConvertBonded):
    pass
